/*
* Copyright (c) 2003, University of Maryland 
* Copyright (c) 2003, Center for Satellite and Hybrid Communication Networks 
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions 
* are met:
*  
*   * Redistributions of source code must retain the above copyright notice, 
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the University of Maryland/Center for Satellite
*     and Hybrid Communication Networks nor the names of its 
*     contributors may be used to endorse or promote products derived from 
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
* POSSIBILITY OF SUCH DAMAGE.
*
*/

/* Authors: 
*     Jonathan Polley (jpolley@isr.umd.edu),
*     Dionysus Blazakis (dblaze@isr.umd.edu),
*     Jonathan McGee (mcgee@isr.umd.edu),
*     Daniel Rusk (ruskd@isr.umd.edu),
*     Karl Seamon (kks@isr.umd.edu),
*     Manish Karir (karir@isr.umd.edu)
*/

#include "../EM/atmega128.h"
#include "opcodes.h"
#include "../EM/debug.h"

char g_tempSubDbg=0;

error_t atmega_op_com( atmega_chip *chip )
{
    uint8_t result;
    if(chip->mode == ATEMU_MODE_DECODE){
        log_opcode(chip, "COM\tR%d\n", chip->reg_index_a);
        return 1;
    }

    result = ~chip->a;
    chip->register_file[chip->reg_index_a] = result;

    atmega_alu_set_logic_flags(chip, result);
    SET_FLAG( *chip->sreg, ATEMU_SREG_C );
    
    chip->cycles += 1;
    chip->pc++;
    return 0;
}

error_t atmega_op_neg( atmega_chip *chip )
{
    uint8_t result;
    uint8_t mask_3 = 1 << 3;
    uint8_t mask_7 = 1 << 7;
    char n, v;
    
    if(chip->mode == ATEMU_MODE_DECODE){
        log_opcode(chip, "NEG\tR%d\n", chip->reg_index_a);
        return 1;
    }

    result = ~chip->a + 1;
    chip->register_file[chip->reg_index_a] = result;
    
    /* set H if there is a borrow from bit 3 */
    if( (result & mask_3) | (chip->a & mask_3) )
    {
        SET_FLAG( *chip->sreg, ATEMU_SREG_H );
    }
    else
    {
        CLR_FLAG( *chip->sreg, ATEMU_SREG_H );
    }
    /* set V if two's complement overflow from implied subtraction from zero */
    if( result == 0x80 )
    {
        SET_FLAG( *chip->sreg, ATEMU_SREG_V );
        v = 0x1;
    }
    else
    {
        CLR_FLAG( *chip->sreg, ATEMU_SREG_V );
        v = 0x0;
    }
    /* set N if MSB of result is set */
    if( result & mask_7 )
    {
        SET_FLAG( *chip->sreg, ATEMU_SREG_N );
        n = 0x1;
    }
    else
    {
        CLR_FLAG( *chip->sreg, ATEMU_SREG_N );
        n = 0x0;
    }
    /* set S for signedness */
    if ( v ^ n )
    {
        SET_FLAG( *chip->sreg, ATEMU_SREG_S );
    }
    else
    {
        CLR_FLAG( *chip->sreg, ATEMU_SREG_S );
    }

    /* set Z if result is 0 */
    /* set C unless result is 0x00 */
    if( result == 0x00 )
    {
        SET_FLAG( *chip->sreg, ATEMU_SREG_Z );
        CLR_FLAG( *chip->sreg, ATEMU_SREG_C );
    }
    else
    {
        SET_FLAG( *chip->sreg, ATEMU_SREG_C );
        CLR_FLAG( *chip->sreg, ATEMU_SREG_Z );
    }
   
    chip->cycles += 1;
    chip->pc++; 
    return 0;
}

error_t atmega_op_dec( atmega_chip *chip )
{
    uint8_t result;
    char n,v;
    uint8_t mask_7 = 1 << 7;
    
    if(chip->mode == ATEMU_MODE_DECODE){
        log_opcode(chip, "DEC\tR%d\n", chip->reg_index_a);
        return 1;
    }
    
    result = chip->a - 1;
    chip->register_file[chip->reg_index_a] = result;

    /* set V if two's complement overflow */
    if( chip->a == 0x80 )
    {
        SET_FLAG( *chip->sreg, ATEMU_SREG_V );
        v = 0x1;
    }
    else
    {
        CLR_FLAG( *chip->sreg, ATEMU_SREG_V );
        v = 0x0;
    }

    /* set N if MSB of the result is set */
    if( result & mask_7 )
    {
        SET_FLAG( *chip->sreg, ATEMU_SREG_N );
        n = 0x1;
    }
    else
    {
        CLR_FLAG( *chip->sreg, ATEMU_SREG_N );
        n = 0x0;
    }

    /* set S for signedness */
    if( n ^ v )
    {
        SET_FLAG( *chip->sreg, ATEMU_SREG_S );
    }
    else
    {
        CLR_FLAG( *chip->sreg, ATEMU_SREG_S );
    }

    /* set Z if result is 0 */
    if( result == 0x00 )
    {
        SET_FLAG( *chip->sreg, ATEMU_SREG_Z );
    }
    else
    {
        CLR_FLAG( *chip->sreg, ATEMU_SREG_Z );
    }
    
    chip->cycles += 1;
    chip->pc++;
    return 0;
}

error_t atmega_op_inc( atmega_chip *chip )
{
    uint8_t result;
    char n,v;
    uint8_t mask_7 = 1 << 7;
    
    if(chip->mode == ATEMU_MODE_DECODE){
        log_opcode(chip, "INC\tR%d\n", chip->reg_index_a);
        return 1;
    }
    
    result = chip->a + 1;
    chip->register_file[chip->reg_index_a] = result;

    /* set V if two's complement overflow */
    if( chip->a == 0x7F )
    {
        SET_FLAG( *chip->sreg, ATEMU_SREG_V );
        v = 0x1;
    }
    else
    {
        CLR_FLAG( *chip->sreg, ATEMU_SREG_V );
        v = 0x0;
    }

    /* set N if MSB of the result is set */
    if( result & mask_7 )
    {
        SET_FLAG( *chip->sreg, ATEMU_SREG_N );
        n = 0x1;
    }
    else
    {
        CLR_FLAG( *chip->sreg, ATEMU_SREG_N );
        n = 0x0;
    }

    /* set S for signedness */
    if( n ^ v )
    {
        SET_FLAG( *chip->sreg, ATEMU_SREG_S );
    }
    else
    {
        CLR_FLAG( *chip->sreg, ATEMU_SREG_S );
    }

    /* set Z if result is 0 */
    if( result == 0x00 )
    {
        SET_FLAG( *chip->sreg, ATEMU_SREG_Z );
    }
    else
    {
        CLR_FLAG( *chip->sreg, ATEMU_SREG_Z );
    }
    
    chip->cycles += 1;
    chip->pc++;
    return 0;
}

error_t atmega_op_mul( atmega_chip *chip )
{
    uint16_t result;
    uint16_t mask_15 = 1 << 15;
    
    if(chip->mode == ATEMU_MODE_DECODE){
        log_opcode(chip, "MUL\tR%d,R%d\n", chip->reg_index_a, chip->reg_index_b);
        return 1;
    }

    result = chip->a * chip->b;

    chip->register_file[1] = HIGH_BYTE_16(result);
    chip->register_file[0] = LOW_BYTE_16(result);
    
    if( result & mask_15 )
    {
        SET_FLAG( *chip->sreg, ATEMU_SREG_C );
    }
    else
    {
        CLR_FLAG( *chip->sreg, ATEMU_SREG_C );
    }
    
    if( result == 0x0000 )
    {
        SET_FLAG( *chip->sreg, ATEMU_SREG_Z );
    }
    else
    {
        CLR_FLAG( *chip->sreg, ATEMU_SREG_Z );
    }
    
    chip->cycles += 2;
    chip->pc++;
    return 0;
}

error_t atmega_op_adiw( atmega_chip *chip )
{
    uint16_t data;
    uint16_t result;
    uint8_t regd_l;
    uint8_t regd_h;
    
    regd_l = chip->reg_index_a;
    regd_h = regd_l + 1;
    
    if(chip->mode == ATEMU_MODE_DECODE){
        log_opcode(chip, "ADIW\tR%d, 0x%02x\t;%d\n", regd_l, chip->b, chip->b);
        return 1;
    }

    data = PACK_INT_16(chip->register_file[regd_h], chip->register_file[regd_l]);

    result = data + chip->b;
    chip->register_file[regd_h] = HIGH_BYTE_16(result);
    chip->register_file[regd_l] = LOW_BYTE_16(result);
    
    
    atmega_alu_set_add_flagsw(chip, result, data);
    chip->cycles += 2;
    chip->pc++;
    
    return 0;
}

error_t atmega_op_sbiw( atmega_chip *chip )
{
    uint16_t data;
    uint16_t result;
    uint8_t regd_l;
    uint8_t regd_h;
    
    regd_l = chip->reg_index_a;
    regd_h = regd_l + 1;
    
    if(chip->mode == ATEMU_MODE_DECODE){
        log_opcode(chip, "SBIW\tR%d, 0x%02x\t;%d\n",regd_l, chip->b, chip->b);
        return 1;
    }

    data = chip->register_file[regd_h];
    data <<= 8;
    data |= chip->register_file[regd_l];

    result = data - chip->b;
    chip->register_file[regd_h] = HIGH_BYTE_16(result);
    chip->register_file[regd_l] = LOW_BYTE_16(result);
    
    
    atmega_alu_set_sub_flagsw(chip, result, data);
    
    chip->cycles += 2;
    chip->pc++;
    
    return 0;
}

error_t atmega_op_sbci( atmega_chip *chip )
{
    uint8_t result;
    uint8_t c, z;

    if(chip->mode == ATEMU_MODE_DECODE){
        log_opcode(chip,  "SBCI\tR%d, 0x%02x\t;%d\n", chip->reg_index_a, chip->b, chip->b);
        return 1;
    }

    c = GET_FLAG( *chip->sreg, ATEMU_SREG_C );
    if( c > 0 )
    {
        c = 1;
    }
    result = chip->a - chip->b - c;
    z = GET_FLAG( *chip->sreg, ATEMU_SREG_Z );
       
    atmega_alu_set_sub_flags(chip, result, chip->a, chip->b);
    
    if ( result != 0x00 )
    {
        CLR_FLAG( *chip->sreg, ATEMU_SREG_Z );
    } else {
        if ( z )
            SET_FLAG( *chip->sreg, ATEMU_SREG_Z );
        else
            CLR_FLAG( *chip->sreg, ATEMU_SREG_Z );
    }

    chip->register_file[chip->reg_index_a] = result;
    chip->cycles += 1;
    chip->pc++;
    
    return 0;
}

error_t atmega_op_subi( atmega_chip *chip )
{
    uint8_t result;
    
    if(chip->mode == ATEMU_MODE_DECODE){
        log_opcode(chip,  "SUBI\tR%d, 0x%02x\t;%d\n", chip->reg_index_a, chip->b, chip->b );
        return 1;
    }

    result=chip->a - chip->b;
    chip->register_file[chip->reg_index_a]=result;

    atmega_alu_set_sub_flags(chip, result, chip->a, chip->b);
    
    chip->cycles+=1;
    chip->pc++;

    return 0;
}

error_t atmega_op_ori( atmega_chip *chip )
{
    uint8_t result;

    if(chip->mode == ATEMU_MODE_DECODE){
        log_opcode(chip, "ORI\tR%d, 0x%02x\t;%d\n", chip->reg_index_a, chip->b, chip->b );
        return 1;
    }

    result = chip->a | chip->b;

    atmega_alu_set_logic_flags(chip, result);

    chip->register_file[chip->reg_index_a] = result;

    chip->cycles += 1;
    chip->pc++;

    return 0;
}

error_t atmega_op_andi( atmega_chip *chip )
{
    uint8_t result;

    if(chip->mode == ATEMU_MODE_DECODE){
        log_opcode(chip, "ANDI\tR%d, 0x%02x\t;%d\n", chip->reg_index_a, chip->b, chip->b);
        return 1;
    }

    result = chip->a & chip->b;

    atmega_alu_set_logic_flags(chip, result);
    
    chip->register_file[chip->reg_index_a] = result;
    
    chip->cycles += 1;
    chip->pc++;
    
    return 0;
}

error_t atmega_op_sub( atmega_chip *chip )
{
    uint8_t result;
    
    if(chip->mode == ATEMU_MODE_DECODE){
        log_opcode(chip,  "SUB\tR%d,R%d\n", chip->reg_index_a, chip->reg_index_b );
        return 1;
    }
    result = chip->a - chip->b;
    
    atmega_alu_set_sub_flags(chip, result, chip->a, chip->b);
    
    chip->register_file[chip->reg_index_a] = result;
    
    chip->cycles += 1;
    chip->pc++;

    return 0;
}

error_t atmega_op_adc( atmega_chip *chip )
{
    uint8_t result;
    uint8_t c;
    
    if(chip->mode == ATEMU_MODE_DECODE){
        log_opcode(chip, "ADC\tR%d,R%d\n", chip->reg_index_a, chip->reg_index_b);
        return 1;
    }

    c = GET_FLAG( *chip->sreg, ATEMU_SREG_C );
    if( c > 0 )
    {
        c = 1;
    }
    
    result = chip->a + chip->b + c;
    
    atmega_alu_set_add_flags(chip, result, chip->a, chip->b);
    
    chip->register_file[chip->reg_index_a] = result;
    
    chip->cycles += 1;
    chip->pc++;
    return 0;
}

error_t atmega_op_and( atmega_chip *chip )
{
    uint8_t result = 0;
    
    if(chip->mode == ATEMU_MODE_DECODE){
        log_opcode(chip, "AND\tR%d,R%d\n", chip->reg_index_a, chip->reg_index_b);
        return 1;
    }
    
    result = chip->a & chip-> b;
    
    atmega_alu_set_logic_flags(chip, result);
    
    chip->register_file[chip->reg_index_a] = result;

    chip->cycles += 1;
    chip->pc++;
    
    return 0;
}

error_t atmega_op_eor( atmega_chip *chip )
{
    uint8_t value = 0;

    if(chip->mode == ATEMU_MODE_DECODE){
        log_opcode(chip,  "EOR\tR%d, R%d\n", chip->reg_index_a, chip->reg_index_b );
        return 1;
    }

    value = chip->a ^ chip->b;

    atmega_alu_set_logic_flags( chip, value );
    
    chip->register_file[chip->reg_index_a] = value;

    chip->cycles += 1;
    chip->pc++;

    return 0;
}

error_t atmega_op_or( atmega_chip *chip )
{
    uint8_t result = 0;

    if(chip->mode == ATEMU_MODE_DECODE){
        log_opcode(chip,  "OR\tR%d, R%d\n", chip->reg_index_a, chip->reg_index_b );
        return 1;
    }

    result = chip->a | chip->b;

    atmega_alu_set_logic_flags( chip, result);
    
    chip->register_file[chip->reg_index_a] = result;

    chip->cycles += 1;
    chip->pc++;

    return 0;
}

error_t atmega_op_sbc( atmega_chip *chip )
{
    uint8_t result;
    uint8_t c, z;

    if(chip->mode == ATEMU_MODE_DECODE){
        log_opcode(chip,  "SBC\tR%d,R%d\n", chip->reg_index_a, chip->reg_index_b);
        return 1;
    }

    c = GET_FLAG( *chip->sreg, ATEMU_SREG_C );
    if( c > 0 )
    {
        c = 1;
    }
    
    result = chip->a - chip->b - c;

    z = GET_FLAG( *chip->sreg, ATEMU_SREG_Z );

    atmega_alu_set_sub_flags(chip, result, chip->a, chip->b);
    
    if ( result != 0x00 )
    {
        CLR_FLAG( *chip->sreg, ATEMU_SREG_Z );
    } else {
        if ( z )
            SET_FLAG( *chip->sreg, ATEMU_SREG_Z );
        else
            CLR_FLAG( *chip->sreg, ATEMU_SREG_Z );
    }

    chip->register_file[chip->reg_index_a] = result;

    chip->cycles += 1;
    chip->pc++;

    return 0;
}

error_t atmega_op_add( atmega_chip *chip )
{
    uint8_t result;
          
    if(chip->mode == ATEMU_MODE_DECODE){
        log_opcode(chip, "ADD\tR%d,R%d\n", chip->reg_index_a, chip->reg_index_b);        
        return 1;
    }

    result = chip->a + chip->b;                        
    chip->register_file[chip->reg_index_a] = result;
    
    atmega_alu_set_add_flags(chip, result, chip->a, chip->b);

    chip->cycles += 1;
    chip->pc++;
    return 0;
}

error_t atmega_op_muls( atmega_chip *chip )
{
    int16_t result;
    uint16_t mask_15 = 1 << 15;
    
    if(chip->mode == ATEMU_MODE_DECODE){
        log_opcode(chip, "MULS\tR%d,R%d\n", chip->reg_index_a, chip->reg_index_b);
        return 1;
    }

    result = ((int8_t)chip->a) * ((int8_t)chip->b);

    chip->register_file[1] = HIGH_BYTE_16(result);
    chip->register_file[0] = LOW_BYTE_16(result);
    
    if( result & mask_15 )
    {
        SET_FLAG( *chip->sreg, ATEMU_SREG_C );
    }
    else
    {
        CLR_FLAG( *chip->sreg, ATEMU_SREG_C );
    }
    
    if( result == 0x0000 )
    {
        SET_FLAG( *chip->sreg, ATEMU_SREG_Z );
    }
    else
    {
        CLR_FLAG( *chip->sreg, ATEMU_SREG_Z );
    }

    chip->cycles += 2;
    chip->pc++;
    return 0;
}

error_t atmega_op_mul_group( atmega_chip *chip )
{
    uint8_t mask_4 = 1 << 3;
    error_t result;

    if( !(chip->bh_nibble & mask_4) && (chip->bl_nibble & mask_4))
    {
        result = atmega_op_fmul( chip );
    }
    else if( (chip->bh_nibble & mask_4) && !(chip->bl_nibble & mask_4))
    {
        result = atmega_op_fmuls( chip );
    }
    else if( (chip->bh_nibble & mask_4) && (chip->bl_nibble & mask_4))
    {
        result = atmega_op_fmulsu( chip );
    }
    else if( !(chip->bh_nibble & mask_4) && !(chip->bl_nibble & mask_4))
    {
        result = atmega_op_mulsu( chip );
    }else{
        log_error(chip,  "Unknown instruction\n" );
        result = -1;
    }

    return result;
}

error_t atmega_op_fmul( atmega_chip *chip ) 
{
    uint16_t result;
    uint16_t mask_15 = 1 << 15;
    
    if(chip->mode == ATEMU_MODE_DECODE){
        log_opcode(chip, "FMUL\tR%d,R%d\n", chip->reg_index_a, chip->reg_index_b);
        return 1;
    }
    
    result = (uint16_t)chip->a * (uint16_t)chip->b;
    
    if( result & mask_15 )
    {
        SET_FLAG( *chip->sreg, ATEMU_SREG_C );
    }
    else
    {
        CLR_FLAG( *chip->sreg, ATEMU_SREG_C );
    }
    
    if( result == 0x0000 )
    {
        SET_FLAG( *chip->sreg, ATEMU_SREG_Z );
    }
    else
    {
        CLR_FLAG( *chip->sreg, ATEMU_SREG_Z );
    }
    
    result <<=1;
    chip->register_file[1] = HIGH_BYTE_16(result);
    chip->register_file[0] = LOW_BYTE_16(result);
    
    chip->cycles += 2;
    chip->pc++;

    return 0;
}

error_t atmega_op_fmuls( atmega_chip *chip )
{
    int16_t result;
    uint16_t mask_15 = 1 << 15;
    
    if(chip->mode == ATEMU_MODE_DECODE){
        log_opcode(chip, "FMULS\tR%d,R%d\n", chip->reg_index_a, chip->reg_index_b);
        return 1;
    }
    
    result = (int16_t)((int8_t)chip->a) * (int16_t)((int8_t)chip->b);
    
    if( result & mask_15 )
    {
        SET_FLAG( *chip->sreg, ATEMU_SREG_C );
    }
    else
    {
        CLR_FLAG( *chip->sreg, ATEMU_SREG_C );
    }
    
    if( result == 0x0000 )
    {
        SET_FLAG( *chip->sreg, ATEMU_SREG_Z );
    }
    else
    {
        CLR_FLAG( *chip->sreg, ATEMU_SREG_Z );
    }
    
    result <<=1;
    chip->register_file[1] = HIGH_BYTE_16(result);
    chip->register_file[0] = LOW_BYTE_16(result);
    
    chip->cycles += 2;
    chip->pc++;

    return 0;
}

error_t atmega_op_fmulsu( atmega_chip *chip )
{
    int16_t result;
    uint16_t mask_15 = 1 << 15;
    
    if(chip->mode == ATEMU_MODE_DECODE){
        log_opcode(chip, "FMULSU\tR%d,R%d\n", chip->reg_index_a, chip->reg_index_b);
        return 1;
    }
    
    result = (int16_t)((int8_t)chip->a) * (uint16_t)chip->b;
    
    if( result & mask_15 )
    {
        SET_FLAG( *chip->sreg, ATEMU_SREG_C );
    }
    else
    {
        CLR_FLAG( *chip->sreg, ATEMU_SREG_C );
    }
    
    if( result == 0x0000 )
    {
        SET_FLAG( *chip->sreg, ATEMU_SREG_Z );
    }
    else
    {
        CLR_FLAG( *chip->sreg, ATEMU_SREG_Z );
    }
    
    result <<=1;
    chip->register_file[1] = HIGH_BYTE_16(result);
    chip->register_file[0] = LOW_BYTE_16(result);
    
    chip->cycles += 2;
    chip->pc++;

    return 0;
}

error_t atmega_op_mulsu( atmega_chip *chip )
{
    int16_t result;
    uint16_t mask_15 = 1 << 15;
    
    if(chip->mode == ATEMU_MODE_DECODE){
        log_opcode(chip, "MULSU\tR%d,R%d\n", chip->reg_index_a, chip->reg_index_b);
        return 1;
    }
    
    result = (int16_t)((int8_t)chip->a) * (uint16_t)chip->b;
    
    chip->register_file[1] = HIGH_BYTE_16(result);
    chip->register_file[0] = LOW_BYTE_16(result);
    
    if( result & mask_15 )
    {
        SET_FLAG( *chip->sreg, ATEMU_SREG_C );
    }
    else
    {
        CLR_FLAG( *chip->sreg, ATEMU_SREG_C );
    }
    
    if( result == 0x0000 )
    {
        SET_FLAG( *chip->sreg, ATEMU_SREG_Z );
    }
    else
    {
        CLR_FLAG( *chip->sreg, ATEMU_SREG_Z );
    }

    chip->cycles += 2;
    chip->pc++;
    return 0;
}
