/*
* Copyright (c) 2003, University of Maryland 
* Copyright (c) 2003, Center for Satellite and Hybrid Communication Networks 
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions 
* are met:
*
*   * Redistributions of source code must retain the above copyright notice, 
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the University of Maryland/Center for Satellite
*     and Hybrid Communication Networks nor the names of its 
*     contributors may be used to endorse or promote products derived from 
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
* POSSIBILITY OF SUCH DAMAGE.
*
*/

/* Authors: 
*     Jonathan Polley (jpolley@isr.umd.edu),
*     Dionysus Blazakis (dblaze@isr.umd.edu),
*     Jonathan McGee (mcgee@isr.umd.edu),
*     Daniel Rusk (ruskd@isr.umd.edu),
*     Karl Seamon (kks@isr.umd.edu),
*     Manish Karir (karir@isr.umd.edu)
*/

#include "led.h"
/*
* This module implements the minimum necessary functionality of the
* Atmega128 Port A interface to use the Mica2 LEDs in a fashion compatible
* with TinyOS.  Other, legal methods of driving the LEDs (such as using
* alternate pin functions) will not work.  This device requires use of the
* following registers:
*
* DIRECTION - Port A Data Direction register, requires specific bit of LED
*    Bind to DDRA, mask 0x01 (Yellow), 0x02 (Green), or 0x04 (Red)
* PORT - Port A Output Register, requires specific bit of LED
*    Bind to PORTA, same mask as DIRECTION and PIN
* PIN - Port A Input Register, requires specific bit of LED
*    Bind to PINA, same mask as DIRECTION and PORT
*
* This device requires no clocking or interrupts
*
* Note about initial flicker:  On start-up, you may notice the LEDs turn
*    on and then back off in the course of a few op-codes.  This is a side
*    effect of TinyOS start-up.  On reset, the led's port is initialized to
*    an input with an output value of 0.  When TinyOS starts up, it will first
*    turn the port to an output before setting the output bit, so the LED will
*    be turned on when the output is set and then turned back off once the
*    bit is set.  This is typically in a matter of microseconds and cannot be
*    observed on a real device.
*/

/* Local Statics */
static void led_destroy_device(void *device);
static uint8_t led_direction_read(void *dev);
static void led_direction_write(void *dev, uint8_t value);
static uint8_t led_port_read(void *dev);
static void led_port_write(void *dev, uint8_t value);
static uint8_t led_pin_read(void *dev);
static void led_pin_write(void *dev, uint8_t value);

/*
* I/O Manager interface
*/

device_obj* create_device(char *name) {
    led_device *new_dev;
	
    new_dev = (led_device*)malloc(sizeof(led_device));
    if (new_dev == NULL) {
        return NULL;
    }
	
    memset(new_dev, 0, sizeof(led_device));
    new_dev->device.next = NULL;
    new_dev->device.destroy = led_destroy_device;
    new_dev->device.dev = new_dev;
	
    new_dev->reg_ddra.read = led_direction_read;
    new_dev->reg_ddra.write = led_direction_write;
    new_dev->reg_ddra.dev = new_dev;
	
    new_dev->reg_porta.read = led_port_read;
    new_dev->reg_porta.write = led_port_write;
    new_dev->reg_porta.dev = new_dev;
	
    new_dev->reg_pina.read = led_pin_read;
    new_dev->reg_pina.write = led_pin_write;
    new_dev->reg_pina.dev = new_dev;
	
    new_dev->porta_out = 0;
    new_dev->porta_ddr = 0;
    new_dev->ison = 0;
	
    return &new_dev->device;
}

register_obj *add_register(void *dev, char *name) {
    led_device *led = (led_device*)dev;
	
    if(strcasecmp(name, "DIRECTION") == 0) {
        return &(led->reg_ddra);
    } else if(strcasecmp(name, "PORT") == 0) {
        return &(led->reg_porta);
    } else if(strcasecmp(name, "PIN") == 0) {
        return &(led->reg_pina);
    }
	
    return NULL;
}

/*
* Port I/O Callbacks
*/

static void led_destroy_device(void *dev) {
    led_device *led = (led_device*)dev;
    free(led);
}

static uint8_t led_direction_read(void *dev) {
    led_device *led = (led_device*)dev;
    return led->porta_ddr;
}

static void led_direction_write(void *dev, uint8_t value) {
    led_device *led = (led_device*)dev;
    led->porta_ddr = value;
    /* Forward to write so LED will toggle properly on DDR changes */
    led_port_write(dev, led->porta_out);
}

static uint8_t led_port_read(void *dev) {
    led_device *led = (led_device*)dev;
    return led->porta_out;
}

static void led_port_write(void *dev, uint8_t value) {
    led_device *led = (led_device*)dev;
    unsigned newon;
	
    /* LED is on if port is configured as an output and a 0 is written */
    newon = (!value && led->porta_ddr);
	
    /* Notify on changes */
    if(newon != led->ison) {
        if(newon) {
            led->device.device_event(&led->device, "on\n");
        } else {
            led->device.device_event(&led->device, "off\n");
        }
        led->ison = newon;
    }
	
    led->porta_out = value;
}

static uint8_t led_pin_read(void *dev) {
    led_device *led = (led_device*)dev;
    /* If output, return PORT; otherwise, return 1 (consider the LED a weak
	* pull-up) */
    if(led->porta_ddr) {
        return led->porta_out;
    } else {
        return 1;
    }
}

static void led_pin_write(void *dev, uint8_t value) {
    /* Silly user, writing to what is only an input */
}
