/*
* Copyright (c) 2003, University of Maryland 
* Copyright (c) 2003, Center for Satellite and Hybrid Communication Networks 
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions 
* are met:
*
*   * Redistributions of source     code must retain the above copyright notice, 
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the University of Maryland/Center for Satellite
*     and Hybrid Communication Networks nor the names of its 
*     contributors may be used to endorse or promote products derived from 
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
* POSSIBILITY OF SUCH DAMAGE.
*
*/

/* Authors: 
*     Jonathan Polley (jpolley@isr.umd.edu),
*     Dionysus Blazakis (dblaze@isr.umd.edu),
*     Jonathan McGee (mcgee@isr.umd.edu),
*     Daniel Rusk (ruskd@isr.umd.edu),
*     Karl Seamon (kks@isr.umd.edu),
*     Manish Karir (karir@isr.umd.edu)
*/

/*
* Modifications:
*
* Copyright (c) 2004, The Hong Kong University of Science and Technology (HKUST)
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions
* are met:
*
*   * Redistributions of source code must retain the above copyright notice,
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the university nor the names of its
*     contributors may be used to endorse or promote products derived from
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
* POSSIBILITY OF SUCH DAMAGE.
*
*/


/*

  * Modification author: Hejun Wu (vmnet@cs.ust.hk)
  
********************************************************************************/

#include <assert.h>
#include <math.h>
#include <dlfcn.h>
#include "radio.h"
/*
* This module implements the radio and required components for the
* Crossbow Mica2 and Mica2 dot.
*
* The MICA2 and MICA2 DOT have a slightly different interface on Port D.
* The presence of the DOT-only registers will modify this device's behavior
* to be that of the DOT.
*
* Registers:
*
* DDRA - Data Direction Register for Port A
*    MICA2 binds to DDRD, mask 0x40
* PORTA - Output Register for Port A
*    MICA2 binds to PORTD, mask 0x40
* PINA - Input Register for Port A
*    MICA2 binds to PIND, mask 0x40
*
* DDRD - Data Direction Register for Port D
*    MICA2 binds to DDRD, mask 0xD0
*    MICA2 DOT binds to DDRD, mask 0xF8
* PORTD - Output Register for Port D
*    MICA2 binds to PORTD, mask 0xD0
*    MICA2 DOT binds to PORTD, mask 0xF8
* PIND - Input Register for Port D
*    MICA2 binds to PIND, mask 0xD0
*    MICA2 DOT binds to PIND, mask 0xF8
*
* EXTCONTROL - External Interrupt Control Register
*    MICA2 DOT binds to EICRA, mask 0xC0
* EXTMASK - External Interrupt Mask Register
*    MICA2 DOT binds to EIMSK, mask 0x08
* EXTFLAG - External Interrupt Flag Register
*    MICA2 DOT binds to EIFR, mask 0x08
*
* Interrupts:
*
* INT3 - External Interrupt 3
*    MICA2 DOT binds to 5 ($0008)
*
* Device also requires a CLOCK
*/

/* Local static functions */
static void radio_destroy_device(void* device);
static error_t radio_enter_sleep(void* device, uint8_t sleepmode);
static error_t radio_leave_sleep(void *device);

static uint8_t radio_ddra_read(void *dev);
static void radio_ddra_write(void *dev, uint8_t value);
static uint8_t radio_pina_read(void *dev);
static void radio_pina_write(void *dev, uint8_t value);
static uint8_t radio_porta_read(void *dev);
static void radio_porta_write(void *dev, uint8_t value);

static uint8_t radio_ddrd_read(void *dev);
static void radio_ddrd_write(void *dev, uint8_t value);
static uint8_t radio_pind_read(void *dev);
static void radio_pind_write(void *dev, uint8_t value);
static uint8_t radio_portd_read(void *dev);
static void radio_portd_write(void *dev, uint8_t value);

static error_t radio_spi_slave(void *pvoid, void *device, pin_drive_enum value);
static error_t radio_spi_miso(void *pvoid, void *device, pin_drive_enum value);
static error_t radio_spi_mosi(void *pvoid, void *device, pin_drive_enum value);
static double radio_adc_read(void *pvoid, double avcc);
static error_t radio_cc1000_data(radio_cc1000*, void*, pin_drive_enum,
								 pin_drive_enum);
//////////////////////////////////////////////////////////////////////////////////////////////////////
//The following is for power consumption logging

//char status_radio[MAX_NUM_LOGS]={-1};
//unsigned char strRadioOperations[RADIO_OPERATION_TYPES][MAX_OPERATION_STR_LEN]={"RADIO:SLEEP","RADIO:IDLE","RADIO:TRANSMIT","RADIO:RECEIVE","RADIO:IDLE"};
//////////////////////////////////////////////////////////////////////////////////////////////////////


/*
* Startup functions
*/
device_obj* create_device(char *name) {
    radio_device *new_dev = (radio_device*)malloc(sizeof(radio_device));
    if(new_dev == NULL) {
        return NULL;
    }    
	//VMNet tempory debug+++++++++++++++++++++++++++++
    LOG_CURRENT_FUNCTION("radio_create_device", 1);
    memset(new_dev, 0, sizeof(radio_device));
	
    new_dev->device.destroy = radio_destroy_device;
    new_dev->device.sleep = radio_enter_sleep;
    new_dev->device.wake = radio_leave_sleep;
    new_dev->device.dev = new_dev;
	
    new_dev->reg_ddra.read = radio_ddra_read;
    new_dev->reg_ddra.write = radio_ddra_write;
    new_dev->reg_ddra.dev = new_dev;
    new_dev->reg_porta.read = radio_porta_read;
    new_dev->reg_porta.write = radio_porta_write;
    new_dev->reg_porta.dev = new_dev;
    new_dev->reg_pina.read = radio_pina_read;
    new_dev->reg_pina.write = radio_pina_write;
    new_dev->reg_pina.dev = new_dev;
	
    new_dev->reg_ddrd.read = radio_ddrd_read;
    new_dev->reg_ddrd.write = radio_ddrd_write;
    new_dev->reg_ddrd.dev = new_dev;
    new_dev->reg_portd.read = radio_portd_read;
    new_dev->reg_portd.write = radio_portd_write;
    new_dev->reg_portd.dev = new_dev;
    new_dev->reg_pind.read = radio_pind_read;
    new_dev->reg_pind.write = radio_pind_write;
    new_dev->reg_pind.dev = new_dev;
    new_dev->rssinoise = 0;
    new_dev->spi = NULL;
    
    new_dev->airwave = air_create_device();
    if(new_dev->airwave == NULL) {
        free(new_dev);
        return NULL;
    }
	
    new_dev->cc1000 = cc1000_create_device(new_dev->airwave, & new_dev->device);
    if(new_dev->cc1000 == NULL) {
        air_destroy_device(new_dev->airwave);
        free(new_dev);
        return NULL;
    }
    if(cc1000_data_setcallback(new_dev->cc1000, radio_cc1000_data,
		new_dev) < 0) {
        cc1000_destroy_device(new_dev->cc1000);
        air_destroy_device(new_dev->airwave);
        free(new_dev);
        return NULL;
    }
	
    return &(new_dev->device);
}

register_obj *add_register(void *dev, char *name) {
    radio_device *radio = (radio_device*)dev;  
	//VMNet temporary++++++++++++++++++++++++++++++++++++++++++
	LOG_CURRENT_FUNCTION("radio_add_register", 1);
    if(strcasecmp(name, "DDRD") == 0) {
		LOG_CURRENT_FUNCTION("radio_add_register", 2);
        return &(radio->reg_ddrd);
    } else if(strcasecmp(name, "PORTD") == 0) {
		LOG_CURRENT_FUNCTION("radio_add_register", 3);
        return &(radio->reg_portd);
    } else if(strcasecmp(name, "PIND") == 0) {
		LOG_CURRENT_FUNCTION("radio_add_register", 4);
        return &(radio->reg_pind);
    } else if(strcasecmp(name, "EXTCONTROL") == 0) {
		LOG_CURRENT_FUNCTION("radio_add_register", 5);
        radio->device.device_debug(&radio->device, V_DEBUG,
			"Request for MICA2 DOT register, not currently supported.\n");
        return NULL;
    } else if(strcasecmp(name, "EXTMASK") == 0) {
		LOG_CURRENT_FUNCTION("radio_add_register", 6);
        radio->device.device_debug(&radio->device, V_DEBUG,
			"Request for MICA2 DOT register, not currently supported.\n");
        return NULL;
    } else if(strcasecmp(name, "EXTFLAG") == 0) {
		LOG_CURRENT_FUNCTION("radio_add_register", 7);
        radio->device.device_debug(&radio->device, V_DEBUG,
			"Request for MICA2 DOT register, not currently supported.\n");
        return NULL;
    }
	LOG_CURRENT_FUNCTION("radio_add_register", 8);
    return NULL;
}

interrupt_obj* add_interrupt(void *dev, char *name) {
    radio_device *radio = (radio_device*)dev;
	LOG_CURRENT_FUNCTION("radio_add_interrupt",1);
    if(strcasecmp(name, "INT3") == 0) {
		LOG_CURRENT_FUNCTION("radio_add_interrupt",2);
        radio->device.device_debug(&radio->device, V_DEBUG,
			"Request for MICA2 DOT interrupt, not currently supported.\n");
        return NULL;
    }
	
    return NULL;
}

error_t set_param(void *dev, char* key, char *value) {
    radio_device *radio = (radio_device*)dev;
    char *tok;
    double x = 0, y = 0, z = 0;
	LOG_CURRENT_FUNCTION("radio_set_param",1);
    if(strcasecmp(key, "POSITION") == 0) {
        /* Just using the string tokenizer */
        tok = strtok(value, ",");
        if(tok) {
            x = strtod(tok, NULL);
            tok = strtok(NULL, ",");
            if(tok) {
                y = strtod(tok, NULL);
                tok = strtok(NULL, ",");
                if(tok) {
                    z = strtod(tok, NULL);
                }
            }
        }
		
        air_set_position(radio->airwave, x, y, z);
        return 0;
    } else if(strcasecmp(key, "RADIONOISE") == 0) {
        /* Noise level to calculate bit errors */
		//VMNet temporary debug++++++++++++++++++++++++++++++++++++++++++
		LOG_CURRENT_FUNCTION("radio_set_param",1);
        radio->airwave->noiselevel = pow(10.0, (double)strtod(value, NULL) / 10.0);
    } else if(strcasecmp(key, "ATTENUATE") == 0) {
		LOG_CURRENT_FUNCTION("radio_set_param",1);
        /* How much to communicate signal when transmitting */
        radio->airwave->attenuate = pow(10.0, strtod(value, NULL) / 10.0);
    } else if(strcasecmp(key, "RSSINOISE") == 0) {
		LOG_CURRENT_FUNCTION("radio_set_param",1);
        /* Standard deviation (in Volts) of noise to add to RSSI */
        radio->rssinoise = strtod(value, NULL);
    }
	
    return -1;
}

error_t set_device(void *dev, char *name, device_obj *otherdev) {
    radio_device *radio = (radio_device*)dev;
    spi_set_func pinsetfunc;
	LOG_CURRENT_FUNCTION("radio_set_device",1);
    if(strcasecmp(name, "SPI") == 0) {
		LOG_CURRENT_FUNCTION("radio_set_device",2);
        radio->spi = (spi_device*)otherdev->dev;
        radio->spi_drive_slave =
            (spi_drive_func)dlsym(otherdev->dlhandle, "spi_drive_slave");
        radio->spi_drive_miso =
            (spi_drive_func)dlsym(otherdev->dlhandle, "spi_drive_miso");
        radio->spi_drive_mosi =
            (spi_drive_func)dlsym(otherdev->dlhandle, "spi_drive_mosi");
        radio->spi_drive_clk =
            (spi_drive_func)dlsym(otherdev->dlhandle, "spi_drive_clk");
		
        /* Check the callbacks and float slave */    
        if(radio->spi_drive_slave == NULL || radio->spi_drive_clk == NULL ||
			radio->spi_drive_mosi == NULL || radio->spi_drive_miso == NULL) {
            return -1;
        }
        radio->spi_drive_slave(radio->spi, PIN_FLOAT_LOW);
        
        /* Set the various pin callbacks */
        pinsetfunc =
            (spi_set_func)dlsym(otherdev->dlhandle, "spi_set_slave");
        if(pinsetfunc == NULL) {
            return -1;
        }
        pinsetfunc(radio->spi, radio_spi_slave, radio);
		
        pinsetfunc =
            (spi_set_func)dlsym(otherdev->dlhandle, "spi_set_miso");
        if(pinsetfunc == NULL) {
            return -1;
        }
        pinsetfunc(radio->spi, radio_spi_miso, radio);
		
        pinsetfunc =
            (spi_set_func)dlsym(otherdev->dlhandle, "spi_set_mosi");
        if(pinsetfunc == NULL) {
			
            return -1;
        }
        pinsetfunc(radio->spi, radio_spi_mosi, radio);
		LOG_CURRENT_FUNCTION("radio_set_device",3);
        return 0;
    } else if(strcasecmp(name, "ADC") == 0) {
	/* all we need the ADC for is to assign the callback.  So pointer
		* not stored. */
		LOG_CURRENT_FUNCTION("radio_set_device",4);
        adc_device* adc = (adc_device*)otherdev;
        radio->adc_set_source =
            (adc_set_source_func)dlsym(otherdev->dlhandle, "adc_set_source");
        
        if(radio->adc_set_source == NULL){
            return -1;
        }
		
        radio->adc_set_source(adc, 0, radio_adc_read, radio);
        return 0;
    }
	
    return -1;
}

/*
* Basic device callbacks
*/

static void radio_destroy_device(void *dev) {
    radio_device *radio = (radio_device*)dev;
    cc1000_destroy_device(radio->cc1000);
    air_destroy_device(radio->airwave);
    free(radio);
}

static error_t radio_enter_sleep(void *dev, uint8_t sleepmode) {
	
	//VMNet log++++++++++++++++++++++++++++++++++++++++++
	LOG_CURRENT_OPERATION(RADIO_SLEEP,RADIO_MASK);
	return 0;
}

static error_t radio_leave_sleep(void *dev) {
	
    return 0;
}

/*
* Port A interface
*/

static uint8_t radio_ddra_read(void *dev) {
    radio_device *radio = (radio_device*)dev;
	//VMNet log++++++++++++++++++++++++++++++++++++++++++
    LOG_CURRENT_OPERATION(RADIO_CONTROL,RADIO_MASK);
    return radio->porta_ddr;
}

static void radio_ddra_write(void *dev, uint8_t value) {
    radio_device *radio = (radio_device*)dev;
	//VMNet log++++++++++++++++++++++++++++++++++++++++++++++
    LOG_CURRENT_OPERATION(RADIO_CONTROL,RADIO_MASK);
	
    radio->porta_ddr = value;
    if(value != 0) {
        radio->device.device_debug(&radio->device, V_DEBUG,
			"CHP_OUT on Port A (PA6) is input-only, was set to output.\n");
    }
}

static uint8_t radio_pina_read(void *dev) {
    radio_device *radio = (radio_device*)dev;
	//VMNet log++++++++++++++++++++++++++++++++++++++++++++++
    LOG_CURRENT_OPERATION(RADIO_CONTROL,RADIO_MASK);
    return radio->porta_ddr
        ? radio->porta_out
        : !GET_FLAG(cc1000_chpout_read(radio->cc1000), PIN_DRIVE_LOW);
}

static void radio_pina_write(void *dev, uint8_t value) {
    /* Writing to PINA does absolutely nothing */
}

static uint8_t radio_porta_read(void *dev) {
    radio_device *radio = (radio_device*)dev;
	//VMNet log++++++++++++++++++++++++++++++++++++++++++++++
    LOG_CURRENT_OPERATION(RADIO_CONTROL,RADIO_MASK);
    return radio->porta_out;
}

static void radio_porta_write(void *dev, uint8_t value) {
    radio_device *radio = (radio_device*)dev;
	//VMNet log++++++++++++++++++++++++++++++++++++++++++++++
    LOG_CURRENT_OPERATION(RADIO_CONTROL,RADIO_MASK);
    radio->porta_out = value;
}

/*
* Port D interface
*/

static uint8_t radio_ddrd_read(void *dev) {
    radio_device *radio = (radio_device*)dev;
    LOG_CURRENT_OPERATION(RADIO_CONTROL,RADIO_MASK);
	//printf("hahaha got data from radio heheh");
    return radio->portd_ddr;
}
static int tempRadioFlag=0;
static void radio_ddrd_write(void *dev, uint8_t value) {
    radio_device *radio = (radio_device*)dev;
	//VMNet log++++++++++++++++++++++++++++++++++++++++++++++
    LOG_CURRENT_OPERATION(RADIO_CONTROL,RADIO_MASK);
    radio->portd_ddr = value;
    /* Pass to PORTD so if neat tricks are being played */
	//whj remove    radio_portd_write(dev, radio->portd_out);
}

static uint8_t radio_pind_read(void *dev) {
    radio_device *radio = (radio_device*)dev;
    uint8_t dataout;
    pin_drive_enum clock, data, pale;           
	//VMNet log++++++++++++++++++++++++++++++++++++++++++++++
    LOG_CURRENT_OPERATION(RADIO_CONTROL,RADIO_MASK);
	
    /* If the DDR bit is set, we just use PORTD */
    dataout = (radio->portd_ddr & radio->portd_out);
	
    /* Fill in any inputs */
    cc1000_program_read(radio->cc1000, &clock, &data, &pale);
    if(!GET_FLAG(radio->portd_ddr, RADIO_PORTD_PALE)) {
        if(GET_FLAG(pale, PIN_DRIVE_HIGH)) {
            SET_FLAG(dataout, RADIO_PORTD_PALE);
        }
    }
    if(!GET_FLAG(radio->portd_ddr, RADIO_PORTD_PCLK)) {
        if(GET_FLAG(clock, PIN_DRIVE_HIGH)) {
            SET_FLAG(dataout, RADIO_PORTD_PCLK);
        }
    }
    if(!GET_FLAG(radio->portd_ddr, RADIO_PORTD_PDATA)) {
        if(GET_FLAG(data, PIN_DRIVE_HIGH)) {
            SET_FLAG(dataout, RADIO_PORTD_PDATA);
        }
    }
	
    return dataout;
}

static void radio_pind_write(void *dev, uint8_t value) {
    /* Writing to PIND does absolutely nothing */
    radio_device *radio = (radio_device*)dev;
	//VMNet log++++++++++++++++++++++++++++++++++++++++++++++
    LOG_CURRENT_OPERATION(RADIO_CONTROL,RADIO_MASK);
    radio->device.device_debug(& radio->device, V_DEBUG,
		"User wrote to PIND, huh?!\n");
}

static uint8_t radio_portd_read(void *dev) {
    radio_device *radio = (radio_device*)dev;            
	//VMNet log++++++++++++++++++++++++++++++++++++++++++++++
    LOG_CURRENT_OPERATION(RADIO_CONTROL,RADIO_MASK);
    return radio->portd_out;
}

static void radio_portd_write(void *dev, uint8_t value) {
    radio_device *radio = (radio_device*)dev;
    pin_drive_enum clock, data, pale;
	
    /* Set up clock, data, pale for call into CC1000 */
	
	//VMNet log++++++++++++++++++++++++++++++++++++++++++++++
    LOG_CURRENT_OPERATION(RADIO_CONTROL,RADIO_MASK);
    radio->portd_out = value;
    if(GET_FLAG(radio->portd_ddr, RADIO_PORTD_PALE))
    {
        pale = GET_FLAG(value, RADIO_PORTD_PALE)
            ? PIN_DRIVE_HIGH : PIN_DRIVE_LOW;
		LOG_CURRENT_FUNCTION("radio_radio_portd_write",2);
    }
    else {
		LOG_CURRENT_FUNCTION("radio_radio_portd_write",3);
        pale = PIN_FLOAT;
    }
    if(GET_FLAG(radio->portd_ddr, RADIO_PORTD_PCLK))
	{
		LOG_CURRENT_FUNCTION("radio_radio_portd_write",4);
        clock = GET_FLAG(value, RADIO_PORTD_PCLK)
            ? PIN_DRIVE_HIGH : PIN_DRIVE_LOW;
    }
	else {
        clock = PIN_FLOAT;
		LOG_CURRENT_FUNCTION("radio_radio_portd_write",5);
    }
	
    if(GET_FLAG(radio->portd_ddr, RADIO_PORTD_PDATA))
	{
        data = GET_FLAG(value, RADIO_PORTD_PDATA)
            ? PIN_DRIVE_HIGH : PIN_DRIVE_LOW;
		LOG_CURRENT_FUNCTION("radio_radio_portd_write",6);
    }
	else {
		LOG_CURRENT_FUNCTION("radio_radio_portd_write",7);
        data = PIN_FLOAT;
    }
	
    /* Call CC1000 to program the configure regitster*/
    cc1000_program_write(radio->cc1000, clock, data, pale);
}

/*
* Interaction with other hardware
*/

static error_t radio_spi_slave(void *pvoid, void *device,
							   pin_drive_enum value) {
							   /* The SPI has an external pull-down.  If it ever floats the pin, we need
	* to set it back to pull-down. */
    radio_device *radio = (radio_device*)pvoid;
	
    if(value == PIN_FLOAT) {
		LOG_CURRENT_FUNCTION("radio_radio_spi_slave",2);
        return radio->spi_drive_slave((spi_device*)device, PIN_FLOAT_LOW);
    }
	
    return 0;
}

static inline error_t radio_data_recalc(radio_device *radio,
										pin_drive_enum miso, pin_drive_enum mosi, pin_drive_enum dio) {
    /* Complex Dance: CC1000 <-> MOSI <-RRRR-> MISO */
	LOG_CURRENT_FUNCTION("radio_radio_data_recalc",1);
    if(!GET_FLAG(dio, PIN_FLOAT)) {
        /* Check if radio is driving */
		LOG_CURRENT_FUNCTION("radio_radio_data_recalc",2);
        radio->spi_drive_mosi(radio->spi, dio);
        radio->spi_drive_miso(radio->spi, dio|PIN_FLOAT);
    } else if(!GET_FLAG(mosi, PIN_FLOAT)) {
        /* Check if MOSI is driving */
		LOG_CURRENT_FUNCTION("radio_radio_data_recalc",3);
        radio->spi_drive_miso/* whj correct: spi_drive_miso*/(radio->spi, mosi|PIN_FLOAT);
		
        cc1000_data_write(radio->cc1000, mosi);
    } else if(!GET_FLAG(miso, PIN_FLOAT)) {
        /* Check if MISO is driving */
		LOG_CURRENT_FUNCTION("radio_radio_data_recalc",4);
        radio->spi_drive_mosi/* whj correct: spi_drive_mosi*/(radio->spi, miso|PIN_FLOAT);
		
        cc1000_data_write(radio->cc1000, miso|PIN_FLOAT);
    } else {
        /* All are floating, ignoring weak pulls */
		LOG_CURRENT_FUNCTION("radio_radio_data_recalc",5);
		
        radio->spi_drive_miso(radio->spi, PIN_FLOAT);
        radio->spi_drive_mosi(radio->spi, PIN_FLOAT);
        cc1000_data_write(radio->cc1000, PIN_FLOAT);
    }
	
    return 0;
}

static error_t radio_spi_miso(void *pvoid, void *device,
							  pin_drive_enum value) {
	
    radio_device *radio = (radio_device*)pvoid;
    spi_device *spi = (spi_device*)device;
	
    pin_drive_enum data;
	
    cc1000_data_read(radio->cc1000, &data, NULL);
	
    return radio_data_recalc(radio, value, spi_read_mosi(spi), data);
}

static error_t radio_spi_mosi(void *pvoid, void *device,
							  pin_drive_enum value) {
    radio_device *radio = (radio_device*)pvoid;
    spi_device *spi = (spi_device*)device;
	LOG_CURRENT_FUNCTION("radio_radio_spi_mosi",1);
    pin_drive_enum data;
	
	cc1000_data_read(radio->cc1000, &data, NULL);
	
	
    return radio_data_recalc(radio, spi_read_miso(spi), value, data);
}

static double radio_adc_read(void *pvoid, double avcc) {
    radio_device *radio = (radio_device*)pvoid;
	//VMNet log++++++++++++++++++++++++++++++++++++++++++++++
    double rssi = cc1000_rssi_read(radio->cc1000);
    LOG_CURRENT_OPERATION(RADIO_CONTROL,RADIO_MASK);
    /* Add gaussian noise to the RSSI return if requested */
    if(radio->rssinoise) {
		
        rssi += radio->rssinoise * air_gaussian_random();
    }
	
    return rssi;
}

static error_t radio_cc1000_data(radio_cc1000 *cc1000, void *device,
								 pin_drive_enum clock, pin_drive_enum data) {
    radio_device *radio = (radio_device*)device;
	
    /* Drive the clock directly */
    radio->spi_drive_clk(radio->spi, clock);
	
    /* Defer the data */
    //printf("reading data from radio %d \n", data);
    return radio_data_recalc(radio, spi_read_miso(radio->spi),
		spi_read_mosi(radio->spi), data);
}
