/*
* Copyright (c) 2003, University of Maryland 
* Copyright (c) 2003, Center for Satellite and Hybrid Communication Networks 
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions 
* are met:
*
*   * Redistributions of source code must retain the above copyright notice, 
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the University of Maryland/Center for Satellite
*     and Hybrid Communication Networks nor the names of its 
*     contributors may be used to endorse or promote products derived from 
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
* POSSIBILITY OF SUCH DAMAGE.
*
*/

/* Authors: 
*     Jonathan Polley (jpolley@isr.umd.edu),
*     Dionysus Blazakis (dblaze@isr.umd.edu),
*     Jonathan McGee (mcgee@isr.umd.edu),
*     Daniel Rusk (ruskd@isr.umd.edu),
*     Karl Seamon (kks@isr.umd.edu),
*     Manish Karir (karir@isr.umd.edu)
*/
/*
* Modifications:
*
* Copyright (c) 2004, The Hong Kong University of Science and Technology (HKUST)
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions
* are met:
*
*   * Redistributions of source code must retain the above copyright notice,
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the university nor the names of its
*     contributors may be used to endorse or promote products derived from
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
* POSSIBILITY OF SUCH DAMAGE.
*
*/


/*

  * Modification author: Hejun Wu (vmnet@cs.ust.hk)
  
********************************************************************************/

#include <assert.h>
#include <math.h>
#include <dlfcn.h>
#include "radio.h"
#include "../EM/atmega128.h"
/*
* This source file implements the basic core logic for the CC1000 chip.  It
* is abstracted away so that changes to the wiring can be completed easily
* no matter how strange the schematic gets.  See radio.h for the API.
*/

/* Tuner parameters */
/* Assume tuner width of 600 kHz */
static const double g_deltafreq = 300e3;
/* Currently assume reception threshold of -110dBm */
static const double g_threshold = 1e-11;
/* Frequency of the baud rates as defined for crystal of 14.7456MHz
* and a XOSC_FREQ of 0x03 in MODEM0.  The real baud rate comes from
* this_value * (crystal / 14.7456MHz) * (4 / XOSC_FREQ+1) */
static const double g_baudrates[] = {
    600,    /* 0.6 kBaud */
		1200,   /* 1.2 kBaud */
		2400,   /* 2.4 kBaud */
		4800,   /* 4.8 kBaud */
		9600,   /* 9.6 kBaud */
		19200,  /* 19.2, 38.4, or 76.8 kBaud */
		0,      /* Not used */
		0,      /* Not used */
};

/* Declare local static functions */
static error_t cc1000_drive_data(radio_cc1000* radio, pin_drive_enum data,
								 pin_drive_enum clock);
#if 0
/* Not yet implemented */
static error_t cc1000_drive_chpout(radio_cc1000* radio, pin_drive_enum chpout);
#endif
static error_t cc1000_update_freqs(radio_cc1000* radio);

static uint8_t cc1000_config_read(radio_cc1000* radio, uint8_t address);
static error_t cc1000_config_write(radio_cc1000* radio, uint8_t address,
								   uint8_t value);
static error_t cc1000_config_control(radio_cc1000* radio, uint8_t value);
static error_t cc1000_config_program(radio_cc1000* radio, uint8_t address,
									 uint8_t value);
/* Clocks */
static int cc1000_clock_idle(void *device);
static int cc1000_clock_manreceive(void *device);
static int cc1000_clock_mansend(void *device);
static int cc1000_clock_nrzreceive(void *device);
static int cc1000_clock_nrzsend(void *device);
static int cc1000_clock_asyncreceive(void *device);
static int cc1000_clock_asyncsend(void *device);

static int tempdebugflag=0;
/*
* Public Interface
*/
radio_cc1000* cc1000_create_device(radio_air* airwave, device_obj *device) {
    radio_cc1000* new_dev = (radio_cc1000*)malloc(sizeof(radio_cc1000));
    if(new_dev == NULL) {
        return NULL;
    }
	
    /* Initialize structure */
    memset(new_dev, 0, sizeof(radio_cc1000));
    LOG_CURRENT_FUNCTION("radiocc_cc1000_create_device",1);
	
    new_dev->airwave = airwave;
    new_dev->device = device;
    new_dev->clk_tick.tick = NULL;
    new_dev->clk_tick.dev = new_dev;
    new_dev->func_data = NULL;
    new_dev->func_chpout = NULL;
	
    new_dev->config[RADIO_REG_MAIN] = 0xFF;
    new_dev->prgm_data = PIN_FLOAT;
    new_dev->pin_dio = PIN_FLOAT;
    new_dev->pin_dclk = PIN_FLOAT;
    new_dev->pin_chpout = PIN_DRIVE_LOW;
    /* Assuming Mica2 clock of 8MHz initially */
	//    new_dev->freq_cpu = 8E6;//whj change
    /* Assuming Mica2 clock of 14.7MHz initially */
    new_dev->freq_crystal = 14.7456E6;
    new_dev->freq_cpu=new_dev->freq_crystal/2;
    new_dev->air_asyncsend = 0;
	
    new_dev->cfg_oldclock = 1;
    new_dev->cfg_oldale = 1;
	
    /* Give the airwave some sane values */
    air_transmit_release(airwave);
    air_transmit_power(airwave, 0);
	
    return new_dev;
}

error_t cc1000_destroy_device(radio_cc1000 *radio) {
    /* Release the clock, air and return */
	
    device_clrclock(*radio->device, radio->clk_tick);
    air_transmit_release(radio->airwave);
    free(radio);
    return 0;
}              
////////////////////////////////////////////////////////////
//VMNet
//This function is rewrote
//////////////
error_t cc1000_program_write(radio_cc1000 *radio, pin_drive_enum clockin,
							 pin_drive_enum datain, pin_drive_enum palein) {
    /* Decode clock/data/pale into boolean values */
	/* Clock defaults HIGH if floated */
    int clock = !GET_FLAG(clockin, PIN_DRIVE_LOW);
	/* Doesn't really matter what data reads if floated */
    int data = GET_FLAG(datain, PIN_DRIVE_HIGH);
	/* ALE defaults HIGH if floated */
    int ale = !GET_FLAG(palein, PIN_DRIVE_LOW);
	
    /* Check for clock pulse */
    if((clock != radio->cfg_oldclock) && (radio->cfg_active)) {
        /* Clock edge */
        if(clock == 0) {
			
            /* Falling edge, sample data */
            if(radio->cfg_readaddress) {
				
                if(radio->cfg_position < 8) {
                    radio->cfg_address = (radio->cfg_address << 1) | data;
                    radio->cfg_position++;
                } else {
                    /* Clocking too much data */
                    radio->device->device_debug(radio->device, V_DEBUG,
						"Clocking too many address bits.\n");
					
                }
            } else if(radio->cfg_writing) {
                if(radio->cfg_position < 8) {
                    radio->cfg_data = (radio->cfg_data << 1) | data;
                    radio->cfg_position++;
                    if(radio->cfg_position == 8) {
                        /* Last bit, so commit to device and leave mode */
                        radio->cfg_active = 0;
                        cc1000_config_write(radio, radio->cfg_address,
							radio->cfg_data);
                    }
                }
            }
            else {
				/* Rising edge, output data if reading */
				if(!radio->cfg_readaddress && !radio->cfg_writing) {
					if(radio->cfg_position < 8) {
						/* Normal read cycle */
						data = GET_FLAG(radio->cfg_data, 0x80);
						radio->cfg_data <<= 1;
						radio->cfg_position++;
						radio->prgm_data = (data ? PIN_DRIVE_HIGH : PIN_DRIVE_LOW);
					} else {
						/* End of read cycle, release bus */
						radio->cfg_active = 0;
						radio->prgm_data = PIN_FLOAT;
						radio->device->device_debug(radio->device, V_DEBUG,
                            "Program read cycle complete.\n");
					} /*if(radio->cfg_position < 8) {*/
				} /*if(!radio->cfg_readaddress && !radio->cfg_writing) {*/
			}
        }  /*if(clock == 0) {*/
    } /*if((clock != radio->cfg_oldclock) && (radio->cfg_active)) {*/
	
    /* Check for Address Latch Pulse */
    if(ale != radio->cfg_oldale)
    {
        if(ale == 0) {
            /* Falling edge, beginning programming cycle */
			//VMNet ++++rewrite here. cfg_active +++++++++++++++++++++++++++++
            if(!radio->cfg_active) {
				
                radio->device->device_debug(radio->device, V_DEBUG,
					"Beginning programming cycle.\n");
                radio->cfg_active = 1;
                radio->cfg_readaddress = 1;
                radio->cfg_address = 0;
                radio->cfg_position = 0;
				
            } else {
                LOG_CURRENT_FUNCTION("radiocc_cc1000_program_write",70);
                /* User is still in the middle of a programming cycle */
                radio->device->device_debug(radio->device, V_DEBUG,
					"Address Latch before previous cycle complete.\n");
				//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
            }//whj
        } else if(radio->cfg_active && (radio->cfg_position == 8)) {
			LOG_CURRENT_FUNCTION("radiocc_cc1000_program_write",90);
            /* Rising edge, begin clocking data */
            radio->cfg_readaddress = 0;
            radio->cfg_position = 0;
            /* LSB of address is the Read/Write mode select */
            radio->cfg_writing = GET_FLAG(radio->cfg_address, 1);
            radio->cfg_address >>= 1;
            if(!radio->cfg_writing) {
                /* If in read mode, set up first bit */
                radio->cfg_data = cc1000_config_read(radio,
					radio->cfg_address);
                radio->prgm_data = GET_FLAG(radio->cfg_data, 0x80)
                    ? PIN_DRIVE_HIGH : PIN_DRIVE_LOW;
                radio->cfg_data <<= 1;
                radio->cfg_position++;
            } else {
                /* Write mode, prepare ourselves */
                radio->cfg_data = 0;
                radio->device->device_debug(radio->device, V_DEBUG,
					"Beginning program write cycle to 0x%02X\n",
					radio->cfg_address);
            }
        } else {
            /* Either too short of an address or spurious ALE */
            radio->device->device_debug(radio->device, V_DEBUG,
				"Programming address too short.\n");
            radio->cfg_active = 0;
        }
    } /*if(ale != radio->cfg_oldale) {*/
	
    /* Check if wrong direction on data */
    if(radio->cfg_active && (radio->cfg_position != 8)) {
        if(datain == PIN_FLOAT) {
			
            /* Pin is floating (so acting as an output for us) */
            if(radio->cfg_readaddress || radio->cfg_writing) {
                /* There should be data on the line */
				
                radio->device->device_debug(radio->device, V_DEBUG,
					"Program data pin floating when should be driven.\n");
            }
        } else {
			
            /* Pin is being driven (so acting as an input for us) */
            if(!radio->cfg_readaddress && !radio->cfg_writing) {
				LOG_CURRENT_FUNCTION("radiocc_cc1000_program_write",120);
                /* This should be an output */
				
                radio->device->device_debug(radio->device, V_DEBUG,
					"Program data pin driven when should be floating.\n");
            }
        }
    }
	
    /* Record current clock and ale */
    radio->cfg_oldclock = clock;
    radio->cfg_oldale = ale;
    return 0;
}

error_t cc1000_program_read(radio_cc1000 *radio, pin_drive_enum *clockout,
							pin_drive_enum *dataout, pin_drive_enum *paleout) {
    /* We don't drive either clock or pale */
    if(clockout) {
		LOG_CURRENT_FUNCTION("radiocc_cc1000_program_read",10);
        *clockout = PIN_FLOAT;
    }
    if(paleout) {
		LOG_CURRENT_FUNCTION("radiocc_cc1000_program_read",20);
        *paleout = PIN_FLOAT;
    }
	
    /* Data is floated in address, read and off modes */
    if(dataout) {
		LOG_CURRENT_FUNCTION("radiocc_cc1000_program_read",30);
        *dataout = radio->prgm_data;
    }
	
    return 0;
}

error_t cc1000_data_write(radio_cc1000 *radio, pin_drive_enum data) {
    /* Data writes are only legal if DIO is floating */
    LOG_CURRENT_OPERATION(RADIO_CONTROL,RADIO_MASK);
    if(!GET_FLAG(data, PIN_FLOAT) && !GET_FLAG(radio->pin_dio, PIN_FLOAT)) {
        radio->device->device_debug(radio->device, V_DEBUG,
			"Attempt to drive DIO pin when set as output.\n");
        return -1;
    } else if(!GET_FLAG(radio->pin_dio, PIN_FLOAT)) {
		LOG_CURRENT_FUNCTION("radiocc_cc1000_data_write",20);
        /* Ignore, since we are driving DIO */
        return 0;
    } else {
        /* DIO will still be read as floating, but notify the sending core */
        int bit = GET_FLAG(data, PIN_DRIVE_HIGH);
		
        if(bit == radio->air_onwire) {
			LOG_CURRENT_FUNCTION("radiocc_cc1000_data_write",30);
            return 0;
        }
        /* And notify transmission core (if in async mode) */
		
        radio->air_onwire = bit;
        if(radio->air_asyncsend) {
			
            cc1000_clock_asyncsend(radio);
        }
		
        return 0;
    }
}

error_t cc1000_data_read(radio_cc1000 *radio, pin_drive_enum *dataout,
						 pin_drive_enum *clockout) {
	
    /* Simple copy */
    if(dataout != NULL) {
		LOG_CURRENT_FUNCTION("radiocc_cc1000_data_read",10);
        *dataout = radio->pin_dio;
    }
    if(clockout != NULL) {
		LOG_CURRENT_FUNCTION("radiocc_cc1000_data_read",20);
        *clockout = radio->pin_dclk;
    }
	
    LOG_CURRENT_OPERATION(RADIO_CONTROL,RADIO_MASK);
	//    cc1000_drive_data(radio, PIN_FLOAT, PIN_DRIVE_LOW);   //whj add
    return 0;
}

error_t cc1000_data_setcallback(radio_cc1000 *radio,
								cc1000_data_callback func_data, void *device) {
	LOG_CURRENT_FUNCTION("radiocc_cc1000_data_setcallback",10);
    /* Only one callback can be registered */
    if(radio->func_data != NULL) {
        return -1;
    }
	LOG_CURRENT_FUNCTION("radiocc_cc1000_data_setcallback",20);
    radio->func_data = func_data;
    radio->func_data_dev = device;
    return 0;
}

pin_drive_enum cc1000_chpout_read(radio_cc1000 *radio) {
	LOG_CURRENT_FUNCTION("radiocc_cc1000_chpout_read",10);
    /* Trivial */
    return radio->pin_chpout;
}

error_t cc1000_chpout_setcallback(radio_cc1000 *radio,
								  cc1000_chpout_callback func_chpout, void *device) {
    /* Only one callback can be registered */
    if(radio->func_chpout != NULL) {
		LOG_CURRENT_FUNCTION("radiocc_cc1000_chpout_setcallback",10);
        return -1;
    }
    LOG_CURRENT_FUNCTION("radiocc__cc1000_chpout_setcallback",20);
    radio->func_chpout = func_chpout;
    radio->func_chpout_dev = device;
    return 0;
}

double cc1000_rssi_read(radio_cc1000 *radio) {
    double strength;
	
    if(radio->strength <= 0) {
		LOG_CURRENT_FUNCTION("radiocc_cc1000_rssi_read",10);
        return 1.2;
    }
	
    /* We use the log, or dBm, of the signal strength */
    strength = 10 * log10(radio->strength);
	
    /* If stronger than -50dBm, return 0.1V */
    if(strength > -50) {
		LOG_CURRENT_FUNCTION("radiocc_cc1000_rssi_read",20);
        return 0.1;
    }
	
    /* If weaker than -105dBm, return 1.2V */
    if(strength < -105) {
		LOG_CURRENT_FUNCTION("radiocc_cc1000_rssi_read",30);
        return 1.2;
    }
	
    /* We just linearlize the -105dBm to -50dBm region */
    return 0.1 - (strength + 50) * 1.1 / (105 - 50);
}

/*
* Internal Pin I/O abstraction functions
*/

static error_t cc1000_drive_data(radio_cc1000* radio, pin_drive_enum data,
								 pin_drive_enum clock) {
    if((data != radio->pin_dio) || (clock != radio->pin_dclk)) {
        /* Data has changed, notify upper level */
        radio->pin_dio = data;
        radio->pin_dclk = clock;
		
        if(radio->func_data) {
            return radio->func_data(radio, radio->func_data_dev, clock, data);
        }
    }
    return 0;
}

#if 0
/* Not currently implemented */
static error_t cc1000_drive_chpout(radio_cc1000* radio, pin_drive_enum chpout) {
    if(chpout != radio->pin_chpout) {
        /* Data has changed, notify upper level */
        radio->pin_chpout = chpout;
        if(radio->func_chpout) {
            return radio->func_chpout(radio, radio->func_chpout_dev, chpout);
        }
    }
	
    return 0;
}
#endif

/*
* Radio Reset logic
*/

static int cc1000_reset(radio_cc1000 *radio) {
    double frequency, baud;
	
    radio->device->device_debug(radio->device, V_DEBUG,
		"Resetting all memory locations.\n");
    /* Configuration register defaults */
    radio->config[RADIO_REG_FREQ_2A] = 0x75;
    radio->config[RADIO_REG_FREQ_1A] = 0xA0;
    radio->config[RADIO_REG_FREQ_0A] = 0xCB;
    radio->config[RADIO_REG_FREQ_2B] = 0x75;
    radio->config[RADIO_REG_FREQ_1B] = 0xA5;
    radio->config[RADIO_REG_FREQ_0B] = 0x4E;
    radio->config[RADIO_REG_FSEP1] = 0x00;
    radio->config[RADIO_REG_FSEP0] = 0x59;
    radio->config[RADIO_REG_CURRENT] = 0xCA;
    radio->config[RADIO_REG_FRONT_END] = 0x08;
    radio->config[RADIO_REG_PA_POW] = 0x0F;
    radio->config[RADIO_REG_PLL] = 0x10;
    radio->config[RADIO_REG_LOCK] = 0x00;
    radio->config[RADIO_REG_CAL] = 0x05;
    radio->config[RADIO_REG_MODEM2] = 0x96;
    radio->config[RADIO_REG_MODEM1] = 0x67;
    radio->config[RADIO_REG_MODEM0] = 0x24;
    radio->config[RADIO_REG_MATCH] = 0x00;
    radio->config[RADIO_REG_FSCTRL] = 0x01;
    radio->config[RADIO_REG_FSHAPE7] = 0x01;
    radio->config[RADIO_REG_FSHAPE6] = 0x03;
    radio->config[RADIO_REG_FSHAPE5] = 0x06;
    radio->config[RADIO_REG_FSHAPE4] = 0x0A;
    radio->config[RADIO_REG_FSHAPE3] = 0x11;
    radio->config[RADIO_REG_FSHAPE2] = 0x16;
    radio->config[RADIO_REG_FSHAPE1] = 0x1C;
    radio->config[RADIO_REG_FSDELAY] = 0x2F;
    radio->config[RADIO_REG_PRESCALER] = 0x00;
    radio->config[RADIO_REG_TEST6] = 0x10;
    radio->config[RADIO_REG_TEST5] = 0x04;
    radio->config[RADIO_REG_TEST4] = 0x25;
    radio->config[RADIO_REG_TEST3] = 0x04;
    radio->config[RADIO_REG_TEST2] = 0x00;
    radio->config[RADIO_REG_TEST1] = 0x00;
    radio->config[RADIO_REG_TEST0] = 0x00;
	
    /* Clear any calibration and current clock */
    device_clrclock(*radio->device, radio->clk_tick);
    radio->cal_delay = 0;
	
    /* Frequency nonsense */
    radio->freqA = 0x0075A0CB;
    radio->freqB = 0x0075A54E;
	/* Default divider is / 2 */
    radio->freq_reference = radio->freq_crystal / 2;
	/* Calculate current receive frequency */
    frequency = GET_FLAG(radio->config[RADIO_REG_MAIN], RADIO_REG_MAIN_F_REG)
		? radio->freqB : radio->freqA;
    frequency = radio->freq_reference * (frequency + 8192) / 16384;
	
    /* Calculate default baud rate */
    baud = g_baudrates[(radio->config[RADIO_REG_MODEM0] >> 4) & 0x07];
    baud *= (radio->freq_crystal / 14.7456E6);
    baud *= 4.0 / ((radio->config[RADIO_REG_MODEM0] & 0x03) + 1);
    /* The clock delta is half the clock period */
    radio->air_clockdelta = 0x60; //VMNet changed //(int)(radio->freq_cpu / baud / 2);
    radio->air_clockoriginal = radio->air_clockdelta;
	
    /* Clear the airspace of our mess */
    air_transmit_release(radio->airwave);
	/* Defaults to 0dB */
	
    air_transmit_power(radio->airwave, 1);
	/* Set up the tuner */
    air_receive_tuner(radio->airwave, frequency, g_deltafreq, g_threshold);
	
    return 0;
}

/*
* Program IO access
*/

static uint8_t cc1000_config_read(radio_cc1000 *radio, uint8_t address) {
    uint8_t value;
	
    assert(address < 128);
    value = radio->config[address];
    switch(address) {
    case RADIO_REG_CAL:
        /* Set CAL_COMPLETE if currently calibrating and am complete */
        if(value & RADIO_REG_CAL_START) {
            if(radio->cal_delay > 0) {
                CLR_FLAG(value, RADIO_REG_CAL_COMPLETE);
            } else {
                SET_FLAG(value, RADIO_REG_CAL_COMPLETE);
            }
        } else {
            CLR_FLAG(value, RADIO_REG_CAL_COMPLETE);
        }
        break;
    }
    radio->device->device_debug(radio->device, V_DEBUG,
		"Reading a 0x%02X from radio address 0x%02X\n",
		value, address);
    return value;
}

static error_t cc1000_config_write(radio_cc1000 *radio, uint8_t address,
								   uint8_t value) {
    assert(address < 128);                          
	//VMNet log++++++++++++++++++++++++++++++++++++++++++
    LOG_CURRENT_OPERATION(RADIO_CONTROL,RADIO_MASK);
    radio->device->device_debug(radio->device, V_DEBUG,
		"Wrote a 0x%02X to radio address 0x%02X\n",
		value, address);
	
    /* Software should wait 2ms after enabling core oscillator */
    if(radio->core_delay > 0) {
        radio->device->device_debug(radio->device, V_DEBUG,
            "Software did not wait 2ms after enabling the core oscillator!\n");
        radio->core_delay = 0;
    }
	
    /* See if some special processing is necessary */
    if(address == RADIO_REG_MAIN) {
        cc1000_config_control(radio, value);
        radio->config[RADIO_REG_MAIN] = value;
    } else if(radio->cal_delay > 0) {
		radio->device->device_debug(radio->device, V_DEBUG,
			"Attempt to write to register during calibration.\n");
    } else if(radio->stat_inreset) {
		radio->device->device_debug(radio->device, V_DEBUG,
			"Attempt to write to register during reset.\n");
    } else {
        cc1000_config_program(radio, address, value);
        radio->config[address] = value;
    }
    return 0;
}

static error_t cc1000_config_control(radio_cc1000 *radio, uint8_t value) {
    uint8_t oldvalue = radio->config[RADIO_REG_MAIN];
    /* Check for reset */
	//VMNet log +++++++++++++++++++
    LOG_CURRENT_OPERATION(RADIO_CONTROL,RADIO_MASK);
    if(GET_FLAG(value, RADIO_REG_MAIN_RESET_N) == 0) {
		LOG_CURRENT_FUNCTION("radiocc_cc1000_config_control",20);
        if(!radio->stat_inreset) {
            radio->device->device_debug(radio->device, V_DEBUG,
				"Radio entering reset.\n");
        }
        radio->stat_inreset = 1;
        cc1000_reset(radio);
    } else {
        if(radio->stat_inreset) {
            radio->device->device_debug(radio->device, V_DEBUG,
				"Radio leaving reset.\n");
        }
        radio->stat_inreset = 0;
    }
	
    /* See if the user is disabling a clock while in use */
    if(GET_FLAG(value, RADIO_REG_MAIN_CORE_PD)) {
        /* Main clock generator disabled, everything is halted */
        air_transmit_release(radio->airwave);
        device_clrclock(*radio->device, radio->clk_tick);
        /* Display message if was enabled */
        if(!GET_FLAG(oldvalue, RADIO_REG_MAIN_CORE_PD)) {
            radio->device->device_debug(radio->device, V_DEBUG,
				"Core oscillator disabled, radio halted.\n");
        }
        return 0;
    } else if(GET_FLAG(oldvalue, RADIO_REG_MAIN_CORE_PD)) {
        /* Core clock is starting up */
        radio->device->device_debug(radio->device, V_DEBUG,
			"Core oscillator starting up, give it 2 milliseconds.\n");
        radio->core_delay = (int)(2E-3 * radio->freq_cpu);
    }
	
    /* Check for bias generator being enabled */
    if(!GET_FLAG(value, RADIO_REG_MAIN_BIAS_PD)
		&& GET_FLAG(oldvalue, RADIO_REG_MAIN_BIAS_PD)) {
        /* Bias generator being brought on-line */
        radio->device->device_debug(radio->device, V_DEBUG,
			"Bias generator starting up, give it 200 microseconds.\n");
        radio->bias_delay = (int)(200E-6 * radio->freq_cpu);
    }
	
    /* Check for frequency synthesizer being enabled */
    if(!GET_FLAG(value, RADIO_REG_MAIN_FS_PD)
		&& GET_FLAG(oldvalue, RADIO_REG_MAIN_FS_PD)) {
        /* Bias generator being brought on-line */
        radio->device->device_debug(radio->device, V_DEBUG,
			"Frequency synthesizer starting up, give it 250 microseconds.\n");
        radio->fs_delay = (250E-6 * radio->freq_cpu);
    }
	
    /* Let the delay ticker do the magic */
    device_setclock(*radio->device, radio->clk_tick, cc1000_clock_idle);
    return 0;
}

static error_t cc1000_config_program(radio_cc1000* radio, uint8_t address,
									 uint8_t value) {
    uint8_t oldvalue = radio->config[address], temp;
    double power, baud;
	
    /* Names of LOCK_SELECT values */
    static const char* lockselect[] = {
        "Normal",
			"LOCK_CONTINUOUS",
			"LOCK_INSTANT",
			"ALARM_H",
			"ALARM_L",
			"CAL_COMPLETE",
			"IF_OUT",
			"REFERENCE_DIVIDER",
			"TX_PDB",
			"Manchester Violation",
			"RX_PDB",
			"Not defined",
			"Not defined",
			"LOCK_AVG_FILTER",
			"N_DIVIDER Output",
			"F_COMP",
    };                                              
	//VMNet log ++++++++++++++++++++++++++++++++++++++++
    LOG_CURRENT_OPERATION(RADIO_CONTROL,RADIO_MASK);
    switch(address) {
	case RADIO_REG_FREQ_2A:
	case RADIO_REG_FREQ_1A:
	case RADIO_REG_FREQ_0A:
		/* Radio frequency A */
		radio->freqA = ((uint32_t)(radio->config[RADIO_REG_FREQ_2A]) << 16)
			| ((uint32_t)(radio->config[RADIO_REG_FREQ_1A]) << 8)
			| ((uint32_t)(radio->config[RADIO_REG_FREQ_0A]));
		break;
		
	case RADIO_REG_FREQ_2B:
	case RADIO_REG_FREQ_1B:
	case RADIO_REG_FREQ_0B:
		/* Radio frequency B */
		radio->freqB = ((uint32_t)(radio->config[RADIO_REG_FREQ_2B]) << 16)
			| ((uint32_t)(radio->config[RADIO_REG_FREQ_1B]) << 8)
			| ((uint32_t)(radio->config[RADIO_REG_FREQ_0B]));
		break;
		
	case RADIO_REG_FSEP1:
	case RADIO_REG_FSEP0:
		/* Frequency Seperation Control */
		radio->freqSep
			= ((uint16_t)(radio->config[RADIO_REG_FSEP1] & 0x07) << 8)
			| ((uint16_t)(radio->config[RADIO_REG_FSEP0]));
		break;
		
	case RADIO_REG_PA_POW:
	/* Transmit power select
		* TODO: This is pretty rough */
		if(value == 0) {
			radio->device->device_debug(radio->device, V_DEBUG,
				"Output power set to 0 mW\n");
			
			air_transmit_power(radio->airwave, 0); //whj previous 0
		} else {
			if(value < 0x10) {
				LOG_CURRENT_FUNCTION("radiocc_cc1000_config_program",60);
				/* Assuming linear from 2@-19.5 to 15@-6 */
				power = (value - 2.0) * ((-6.0 + 19.5) / (15.0 - 2.0)) - 19.5;
			} else {
				/* Assuming linear from 4@-5 to 16@+5 */
				power = value / 16.0;
				power = (power - 4.0) * ((5.0 - -5.0) / (16.0 - 4.0)) - 5.0;
			}
			
			radio->device->device_debug(radio->device, V_DEBUG,
				"Output power set to %.1f dBm\n", power);
			/* Set the Transmit power to the value in mW */
			
			power = pow(10.0, power / 10.0);
			radio->device->device_debug(radio->device, V_DEBUG,
				"Also known as %.1e mW\n", power);
			air_transmit_power(radio->airwave, power);
		}
		break;
		
	case RADIO_REG_PLL:
		/* Phase-Locked Loop control (including main clock) */
		temp = (value & 0x78) >> 3;
		if(temp < 2) {
			radio->device->device_debug(radio->device, V_DEBUG,
				"Reference divider set to illegal value %u, assuming 2\n",
				temp);
			temp = 2;
		}
		
		radio->freq_reference = radio->freq_crystal / temp;
		if(radio->freq_reference < 1E6) {
			radio->device->device_debug(radio->device, V_DEBUG,
				"Reference frequency below 1MHz, not allowed\n");
		} else {
			radio->device->device_debug(radio->device, V_DEBUG,
				"Reference frequency set to %.2fMHz\n",
				radio->freq_reference / 1E6);
		}
		break;
		
	case RADIO_REG_LOCK:
		/* Display LOCK_SELECT */
		radio->device->device_debug(radio->device, V_DEBUG,
			"LOCK_SELECT set to: %s\n", lockselect[value >> 4]);
		break;
		
	case RADIO_REG_CAL:
		/* Calibration register */
		if(!GET_FLAG(oldvalue, RADIO_REG_CAL_START)
			&& GET_FLAG(value, RADIO_REG_CAL_START)) {
			/* Delay = 34ms / reference frequency (in MHz) */
			double delay = 34 * 1E6 / radio->freq_reference;
			/* See if halved calibration time */
			if(!GET_FLAG(value, RADIO_REG_CAL_WAIT)) {
				delay /= 2;
			}
			/* Warn user and begin wait (clock called at 8MHz) */
			radio->device->device_debug(radio->device, V_DEBUG,
				"Beginning calibration, should take %.1f milliseconds\n",
				delay);
			radio->cal_delay = (uint32_t)(delay * 1E-3 * radio->freq_cpu);
			device_setclock(*radio->device, radio->clk_tick,
				cc1000_clock_idle);
		} else if(GET_FLAG(oldvalue, RADIO_REG_CAL_START)
			&& !GET_FLAG(value, RADIO_REG_CAL_START)) {
			/* Leaving calibration cycle */
			if(radio->cal_delay > 0) {
				radio->device->device_debug(radio->device, V_DEBUG,
					"Calibration cancelled prior to completion.\n");
			} else {
				radio->device->device_debug(radio->device, V_DEBUG,
					"Software cleared the calibration start bit.\n");
			}
		}
		break;
		
	case RADIO_REG_MODEM0:
		/* Baud Rate */
		baud = g_baudrates[(value >> 4) & 0x07];
		if(baud == 0) {
			/* Illegal baud rate, fall back to 0.6 kBaud */
			radio->device->device_debug(radio->device, V_DEBUG,
				"Software tried to set unused baud rate %d.\n",
				(value >> 4) & 0x07);
			baud = 600;
		}
		baud *= (radio->freq_crystal / 14.7456E6);
		baud *= 4.0 / ((value & 0x03) + 1);
		/* The clock delta is half the clock period */
		radio->air_clockdelta = (int)(radio->freq_cpu / baud / 2);
		
		radio->air_clockoriginal = radio->air_clockdelta;
		radio->device->device_debug(radio->device, V_DEBUG,
			"Baud rate set to %.1f kBaud (actual)\n", baud / 1E3);
		/* Encoding */
		switch(value & RADIO_REG_MODEM0_DATAFORMAT) {
		case RADIO_REG_MODEM0_FORMAT_NRZ:
			radio->device->device_debug(radio->device, V_DEBUG,
				"Modem encoding set to Non-Return to Zero.\n");
			break;
		case RADIO_REG_MODEM0_FORMAT_MAN:
			radio->device->device_debug(radio->device, V_DEBUG,
				"Modem encoding set to Manchester (data rate is half of baud).\n");
			break;
		case RADIO_REG_MODEM0_FORMAT_UART:
			radio->device->device_debug(radio->device, V_DEBUG,
				"Modem encoding set to Asychronous UART (baud will be ignored).\n");
			break;
		default:
			radio->device->device_debug(radio->device, V_DEBUG,
				"Software tried to select unused modem encoding.\n");
			break;
		}
		break;
    }
	
    return 0;
}

static error_t cc1000_update_freqs(radio_cc1000* radio) {
    /* Selected frequency word */
    uint32_t freqWord
        = GET_FLAG(radio->config[RADIO_REG_MAIN], RADIO_REG_MAIN_F_REG)
        ? radio->freqB : radio->freqA;
    double frequency;
    /* Make sure the frequency word is in proper bounds */
    radio->device->device_debug(radio->device, V_DEBUG,
		"Frequency word of %u specified\n", freqWord);
    if(freqWord > 16751615) {
        radio->device->device_debug(radio->device, V_DEBUG,
			"Selected frequency word is out of range, cannot exceed 16751615\n");
        freqWord = 16751615;
    } else if(freqWord < 4194304) {
        radio->device->device_debug(radio->device, V_DEBUG,
			"Selected frequency word is out of range, must be at least 4194304\n");
        freqWord = 4194304;
    }
	
    /* Make sure the frequency seperator is within bounds */
    if((radio->freqSep > 2047)
		&& GET_FLAG(radio->config[RADIO_REG_MAIN], RADIO_REG_MAIN_RXTX)) {
        radio->device->device_debug(radio->device, V_DEBUG,
			"Frequency seperator cannot exceed 2047.\n");
        radio->freqSep = 2047;
    }
	
    /* Calculate frequency */
	
    frequency = radio->freq_reference * (freqWord + 8192) / 16384;
    radio->device->device_debug(radio->device, V_DEBUG,
		"Frequency of %.1fMHz specified\n", frequency / 1E6);
    radio->freqReceive = radio->freqZero = 1E6;
    radio->freqOne = 2E6;
    radio->freqReceive = radio->freqZero = frequency;
    radio->freqOne = frequency
        + (radio->freq_reference * radio->freqSep) / 16384;
	
    return 0;
}

/*
* Clock logic
*/

static int radio_canjoinwire(radio_cc1000 *radio, uint8_t value) {
    /* Look for massive errors */
    if(GET_FLAG(value, RADIO_REG_MAIN_CORE_PD)) {
        radio->device->device_debug(radio->device, V_DEBUG,
			"Cannot continue with core clock disabled!\n");
        return 0;
    }
    if(GET_FLAG(value, RADIO_REG_MAIN_FS_PD)) {
        radio->device->device_debug(radio->device, V_DEBUG,
			"Cannot continue with frequency synthesiser disabled!\n");
        return 0;
    }
    if(GET_FLAG(value, RADIO_REG_MAIN_BIAS_PD)) {
        radio->device->device_debug(radio->device, V_DEBUG,
			"Cannot continue with bias generator disabled!\n");
        return 0;
    }
	
    return 1;
}

static int cc1000_clock_idle(void *device) {
    radio_cc1000 *radio = (radio_cc1000*)device;
    uint8_t control = radio->config[RADIO_REG_MAIN];
    /* Idle does nothing so long as the main clock is off */
    if(GET_FLAG(control, RADIO_REG_MAIN_CORE_PD)) {
        air_transmit_release(radio->airwave);
        device_clrclock(*radio->device, radio->clk_tick);
		//VMNet log ++++++++++++++++++++++++++++++++++++++++
        LOG_CURRENT_OPERATION(RADIO_IDLE,RADIO_MASK);
        return 0;
    }
	
    /* See if we are waiting for the crystal oscillator to stabalize */
    if(radio->core_delay > 0) {
        air_transmit_release(radio->airwave);
        radio->core_delay--;
        if(radio->core_delay == 0) {
            radio->device->device_debug(radio->device, V_DEBUG,
				"Core clock oscillator has stabalized.\n");
        }
        return 0;
    }
	
    /* See if we are waiting for the bias generator to stabalize */
    if(radio->bias_delay > 0) {
        air_transmit_release(radio->airwave);
        radio->bias_delay--;
        if(radio->bias_delay == 0) {
            radio->device->device_debug(radio->device, V_DEBUG,
				"Bias generator has stabalized.\n");
        }
        return 0;
    }
	
    /* See if we are waiting for the clock to stabalize */
    if(radio->fs_delay > 0) {
        air_transmit_release(radio->airwave);
        radio->fs_delay--;
        if(radio->fs_delay == 0) {
            radio->device->device_debug(radio->device, V_DEBUG,
				"Frequency Synthesizer has stabalized.\n");
        }
        return 0;
    }
	
    /* See if we are still in the calibration loop */
    if(radio->cal_delay > 0) {
        air_transmit_release(radio->airwave);
        radio->cal_delay--;
        if(radio->cal_delay == 0) {
            radio->device->device_debug(radio->device, V_DEBUG,
				"Calibration complete.\n");
        }
        return 0;
    }
	
    /* Check if we want to enter transmit or receive */
    if(!GET_FLAG(control, RADIO_REG_MAIN_FS_PD)
		&& !GET_FLAG(control, RADIO_REG_MAIN_BIAS_PD)
		&& !GET_FLAG(control, RADIO_REG_MAIN_CORE_PD)) {
        if(!GET_FLAG(control, RADIO_REG_MAIN_RXTX)) {
            if(!GET_FLAG(control, RADIO_REG_MAIN_RX_PD)) {
                /* Want to receive */
                air_transmit_release(radio->airwave);
                radio->device->device_debug(radio->device, V_DEBUG,
					"Trying to enter receive mode...\n");
                cc1000_update_freqs(radio);
                if(radio_canjoinwire(radio, control)) {
                    switch(radio->config[RADIO_REG_MODEM0]
						& RADIO_REG_MODEM0_DATAFORMAT) {
                    case RADIO_REG_MODEM0_FORMAT_NRZ:
                        /* Non-Return to Zero */
                        device_setclock(*radio->device, radio->clk_tick,
							cc1000_clock_nrzreceive);
                        break;
                    case RADIO_REG_MODEM0_FORMAT_MAN:
                        /* Manchester Encoding */
                        device_setclock(*radio->device, radio->clk_tick,
							cc1000_clock_manreceive);
                        break;
                    case RADIO_REG_MODEM0_FORMAT_UART:
                        /* Asynchronous UART */
                        device_setclock(*radio->device, radio->clk_tick,
							cc1000_clock_asyncreceive);
                        break;
                    default:
                        /* The illegal mode, do nothing */
                        radio->device->device_debug(radio->device, V_DEBUG,
							"Illegal encoding method specified, aborting.\n");
                        break;
                    }
                    radio->air_clockdelta = radio->air_clockoriginal;
                    radio->air_clock = radio->air_clockdelta;
                    radio->air_clockout = 1;
                    radio->air_manchester = 3;
                    radio->sync = 0;
                    radio->sync_state = CC1000_SYNC_LOST;
                    return 0;
                }
            } /* if(!GET_FLAG(control, RADIO_REG_MAIN_RX_PD)) { */
        } else {
            if(!GET_FLAG(control, RADIO_REG_MAIN_TX_PD)) {
                /* Want to transmit */
                radio->device->device_debug(radio->device, V_DEBUG,
					"Trying to enter transmit mode...\n");
                cc1000_update_freqs(radio);
				
                if(radio_canjoinwire(radio, control)) {
					
                    switch(radio->config[RADIO_REG_MODEM0]
						& RADIO_REG_MODEM0_DATAFORMAT) {
                    case RADIO_REG_MODEM0_FORMAT_NRZ:
                        /* Non-Return to Zero */
                        device_setclock(*radio->device, radio->clk_tick,
							cc1000_clock_nrzsend);
                        break;
                    case RADIO_REG_MODEM0_FORMAT_MAN:
                        /* Manchester Encoding */
                        device_setclock(*radio->device, radio->clk_tick,
							cc1000_clock_mansend);
                        break;
                    case RADIO_REG_MODEM0_FORMAT_UART:
                        /* Asynchronous UART */
                        device_setclock(*radio->device, radio->clk_tick,
							cc1000_clock_asyncsend);
                        radio->air_asyncsend = 1;
                        break;
                    default:
                        /* The illegal mode, do nothing */
						
                        radio->device->device_debug(radio->device, V_DEBUG,
							"Illegal encoding method specified, aborting.\n");
                        break;
                    }
					
                    /* Begin with an initial sample */
                    radio->air_clockout = 1;
                    radio->air_clockdelta = radio->air_clockoriginal;
					
                    radio->air_clock = radio->air_clockdelta;
                    radio->air_manchester = 3;
                    cc1000_drive_data(radio, PIN_FLOAT, PIN_DRIVE_LOW);
					
                    return 0;
                }
            } /* if(!GET_FLAG(control, RADIO_REG_MAIN_TX_PD)) { */
        } /* if(GET_FLAG(control, RADIO_REG_MAIN_RXTX)) { */
    }
	
    /* Nothing left to do */
    air_transmit_release(radio->airwave);
    device_clrclock(*radio->device, radio->clk_tick);
    radio->air_asyncsend = 0;
	//VMNet log ++++++++++++++++++++++++++++++++++++++++
    LOG_CURRENT_OPERATION(RADIO_IDLE,RADIO_MASK);
    return 0;
}

/*
* Transmit Logic
*/

static int cc1000_clock_asyncsend(void *device) {
    radio_cc1000 *radio = (radio_cc1000*)device;
    /* The asynchronous UART pretty much just copies the DIO pin
	* directly onto the wire. */
    device_clrclock(*radio->device, radio->clk_tick);
	
	
    return air_transmit(radio->airwave,
		radio->air_onwire ? radio->freqOne : radio->freqZero,0);
}

static int cc1000_clock_nrzsend(void *device) {
    radio_cc1000 *radio = (radio_cc1000*)device;
    /* Sample on falling edge, drive on rising edge, we control the
	* clock. */
	
    if(--radio->air_clock) {
        /* Between clock edges */
		
        return 0;
    }
    radio->air_clock = radio->air_clockdelta;
	
    if(radio->air_clockout) {
		
        /* Falling edge, sample next bit */
        radio->air_transmit = radio->air_onwire;
        radio->air_clockout = 0;
    } else {
        /* Rising edge, drive current bit */
		
        air_transmit(radio->airwave, radio->air_transmit
			? radio->freqOne : radio->freqZero,0);
        radio->air_clockout = 1;
    }
	
    /* Notify device of clock edge */
    return cc1000_drive_data(radio, PIN_FLOAT,
		radio->air_clockout ? PIN_DRIVE_HIGH : PIN_DRIVE_LOW);
}

static int cc1000_clock_mansend(void *device) {
    radio_cc1000 *radio = (radio_cc1000*)device;
    /* The action baud rate is not changed from NRZ, but the data rate is
	* halved.  We handle this by using four stages instead of two.  The
	* current manchester stage is stored in air_manchester and is:
	* 0 - Rising Edge, sample next bit
	* 1 - High Level, drive first chip
	* 2 - Falling Edge
	* 3 - Low Level, drive second chip
    LOG_CURRENT_FUNCTION("radiocc_     */
    if(--radio->air_clock) {
        /* Between clock edges */
		
        return 0;
    }
    radio->air_clock = radio->air_clockdelta;
	
    /* Rotate from 0->1->2->3->0 */
    radio->air_manchester = (radio->air_manchester + 1) & 0x03;
    switch(radio->air_manchester) {
    case 0:
        /* Rising clock edge, sample next bit */
        radio->air_clockout = 1;
        radio->air_transmit = radio->air_onwire;
		
        break;
    case 1:
		//VMNet Manchester
        /* High Level, drive first chip (equal to bit) */
		//              air_transmit(radio->airwave, radio->air_transmit
		//            ? radio->freqZero : radio->freqOne,1);//whj chage to invert
		
        air_transmit(radio->airwave, radio->air_transmit
			? radio->freqOne : radio->freqZero,1); 
        break;
    case 2:
        /* Falling clock edge, nothing special */
		
        radio->air_clockout = 0;
        break;
    case 3:
        /* Low Level, drive second chip (invert of bit) */
		
		//    air_transmit(radio->airwave, radio->air_transmit
		//              ? radio->freqOne : radio->freqZero,1); //whj  chage to equal   set mansend
		
        air_transmit(radio->airwave, radio->air_transmit
			? radio->freqZero : radio->freqOne,1);
        break;
    }
	
    /* Notify device of clock edge */
	
    return cc1000_drive_data(radio, PIN_FLOAT,
		radio->air_clockout ? PIN_DRIVE_HIGH : PIN_DRIVE_LOW);
}

/*
* Receive Logic
*/

static inline int cc1000_receive_sync(radio_cc1000 *radio) {
    int valid, bit;
    double freq, intfreq, seperation;
    static const int bitwait[] = { 11, 22, 43, 86 };
	
    /* Perform the appropriate action based on synchronization stage */
    switch(radio->sync_state) {
    case CC1000_SYNC_NOSIGNAL:
        /* Waiting for signal */
		
        if(air_receive_changed(radio->airwave)) {
			
            valid = air_receive_sense(radio->airwave, &freq, &radio->strength);
            if(valid) {
				
                /* Caught something, try to synchronize */
                radio->sync_state = CC1000_SYNC_FREQLOW;
                radio->sync_chipsleft = radio->sync_chips;
                radio->sync_lastfreq = freq;
                radio->freqOne = radio->freqZero = freq;
                radio->sync_cyclecount = 0;
            }
        }
        return 0;
		
    case CC1000_SYNC_FREQLOW:
        /* Looking for a low->high transition */
        valid = air_receive_sense(radio->airwave, &freq, &radio->strength);
		
        if(!valid) {
            /* Connection lost */
			
            radio->sync_state = CC1000_SYNC_NOSIGNAL;
        } else if(freq < radio->freqZero) {
			
            /* Heading lower */
            radio->freqZero = freq;
            radio->sync_lastfreq = freq;
        } else if(freq < (radio->sync_lastfreq + 1E3)) {
            /* Noise filter */
        } else if(freq > radio->sync_lastfreq) {
			
            /* Low -> High transition */
            radio->sync_state = CC1000_SYNC_FREQHIGH;
            radio->sync_chipsleft--;
            radio->sync_lastfreq = freq;
        }
        return 0;
		
    case CC1000_SYNC_FREQHIGH:
        /* Looking for a high->low transition */
		
        valid = air_receive_sense(radio->airwave, &freq, &radio->strength);
        if(!valid) {
			
            /* Connection lost */
            radio->sync_state = CC1000_SYNC_NOSIGNAL;
        } else if(freq > radio->freqOne) {
			
            /* Heading higher */
            radio->freqOne = freq;
            radio->sync_lastfreq = freq;
        } else if(freq > (radio->sync_lastfreq + 1E3)) {
            /* Noise filter */
        } else if(freq < radio->sync_lastfreq) {
            /* High -> Low transition */
            radio->sync_state = CC1000_SYNC_FREQLOCK;
            radio->sync_chipsleft--;
            radio->sync_lastfreq = freq;
			
        }
        return 0;
		
    case CC1000_SYNC_FREQLOCK:
	/* Looking for final low->high transition (in case first stage
		* caught something after the lowest point */
        valid = air_receive_sense(radio->airwave, &freq, &radio->strength);
		
        if(!valid) {
            /* Connection lost */
			
            radio->sync_state = CC1000_SYNC_NOSIGNAL;
        } else if(freq < radio->freqZero) {
			
            /* Heading lower */
            radio->freqZero = freq;
            radio->sync_lastfreq = freq;
        } else if(freq < (radio->sync_lastfreq + 1E3)) {
            /* Noise filter */
        } else if(freq > radio->sync_lastfreq) {
			
            /* Low -> High transition */
            intfreq = (radio->freqZero + radio->freqOne) / 2
				- radio->freqReceive;
            seperation = radio->freqOne - radio->freqZero;
            /* Check on type of injection */
            if(radio->freqReceive > radio->freqZero) {
                radio->device->device_debug(radio->device, V_DEBUG,
					"Synchronized to high-side injection of frequency"
					" %.1fMHz, intermediate frequency %.0fkHz, frequency"
					" seperation %.1fkHz.  Received data will be"
					" inverted.\n", radio->freqZero / 1E6,
					(radio->freqReceive - radio->freqOne) / 1E3,
					seperation / 1E3);
				LOG_CURRENT_FUNCTION("radiocc_cc1000_receive_sync",155);        
                /* Reverse frequencies for one and zero, schmitt of 2/3 */
                air_receive_frequencies(radio->airwave, radio->freqOne,
					radio->freqZero, 2.0 / 3.0);
            } else {
                radio->device->device_debug(radio->device, V_DEBUG,
					"Synchronized to low-side injection of frequency"
					" %.1fMHz, intermediate %.0fkHz, seperation %.1fkHz.\n",
					radio->freqZero / 1E6,
					(radio->freqOne - radio->freqReceive) / 1E3,
					seperation / 1E3);
                /* Forward frequencies for one and zero, schmitt of 2/3 */
                air_receive_frequencies(radio->airwave, radio->freqZero,
					radio->freqOne, 2.0 / 3.0);
            }
            radio->sync_state = CC1000_SYNC_CHIPWAIT;
            radio->sync_chipsleft--;
			
            /* More notification */
            radio->device->device_debug(radio->device, V_DEBUG,
				"Need %u more chips to synchronize modem.\n",
				radio->sync_chipsleft);
            if((radio->config[RADIO_REG_MODEM0] & RADIO_REG_MODEM0_DATAFORMAT)
				== RADIO_REG_MODEM0_FORMAT_MAN) {
                /* In manchester, use fewer bits for synchronization */
                radio->sync_chipsleft = 20;
            }
            /* Configure tuners */
            air_receive(radio->airwave, &radio->sync_bit, &radio->strength);
            air_receive_tuner(radio->airwave, radio->freqReceive,
				seperation / 2, g_threshold);
        }
        return 0;
		
    case CC1000_SYNC_CHIPWAIT:
        /* Waiting for the next transition */
		
        if(air_receive_changed(radio->airwave)) {
			
            valid = air_receive(radio->airwave, &bit, &radio->strength);
            if(!valid) {
				
                /* Connection lost */
                radio->sync_state = CC1000_SYNC_LOST;
            } else if(bit != radio->sync_bit) {
				
                /* Transition */
                radio->sync_state = CC1000_SYNC_CHIPTIME;
                radio->sync_bitcount = 0;
                radio->sync_cyclecount = 0;
                radio->sync_chipsleft--;
                radio->sync_bit = bit;
            }
        }
        return 0;
		
    case CC1000_SYNC_CHIPTIME:
        /* Timing remaining chips in synchronization to determine data rate */
        radio->sync_cyclecount++;
		
        if(air_receive_changed(radio->airwave)) {
			
            valid = air_receive(radio->airwave, &bit, &radio->strength);
            if(!valid) {
                /* Connection lost */
				
                radio->sync_state = CC1000_SYNC_LOST;
            } else if(bit != radio->sync_bit) {
                /* Transition */
				
                radio->sync_bitcount++;
                radio->sync_bit = bit;
                radio->sync_chipsleft--;
                if(radio->sync_chipsleft == 0) {
                    /* Fully synchronized */
                    radio->sync_state = CC1000_SYNC_LOCKED;
                    radio->sync = 1;
                    radio->device->device_debug(radio->device, V_DEBUG,
						"Detected data rate of %0.1f kilobits per second.\n",
						radio->freq_cpu / 1E3 /
						(radio->sync_cyclecount / radio->sync_bitcount));
                    return 1;
                }
            }
        }
        return 0;
		
    case CC1000_SYNC_LOCKED:
        /* This should probably be an abort() or something drastic */
		
        radio->sync = 1;
        return 1;
		
    default:
        /* Waiting for signal, clean up previous state */
		
		
        radio->sync_state = CC1000_SYNC_NOSIGNAL;
        /* Figure out how many chips to wait for */
        radio->sync_chips = bitwait[(radio->config[RADIO_REG_MODEM1] >> 1)
			& 0x03];
		
        /* Configure tuners */
        air_receive_tuner(radio->airwave, radio->freqReceive, g_deltafreq,
			g_threshold);
        return 0;
    }
	
    return 0;
}

static int cc1000_clock_asyncreceive(void *device) {
    radio_cc1000 *radio = (radio_cc1000*)device;
	
    if(!radio->sync) {
        /* Not synchronized */
        if(!cc1000_receive_sync(radio)) {
			
            return 0;
        }
    }
    LOG_CURRENT_FUNCTION("radiocc_cc1000_clock_asyncreceive",30);
    if(air_receive_changed(radio->airwave)) {
        /* If air activity, update the wire */
        int valid, bit;
        valid = air_receive(radio->airwave, &bit, & radio->strength);
        cc1000_drive_data(radio, PIN_FLOAT,
			(valid && bit) ? PIN_DRIVE_HIGH : PIN_DRIVE_LOW);
        if(!valid) {
            radio->sync = 0;
            radio->sync_state = CC1000_SYNC_LOST;
			
        }
    }
	
    return 0;
}

static int cc1000_clock_nrzreceive(void *device) {
    radio_cc1000 *radio = (radio_cc1000*)device;
	
    if(!radio->sync) {
        if(!cc1000_receive_sync(radio)) {
            /* Not synchronized */
			
            radio->air_clockdelta = radio->air_clockoriginal;
			
        } else {
            /* Synchronization just completed, begin clocking */
			
            radio->air_clockdelta = radio->sync_cyclecount
				/ radio->sync_bitcount / 2;
			
        }
    }
	
    /* Sample on rising edge, drive on falling edge, we control the
	* clock. */
    if(--radio->air_clock) {
		
        /* Between clock edges */
        return 0;
    }
    radio->air_clock = radio->air_clockdelta;
	
    if(radio->air_clockout) {
		
        /* Falling edge */
        cc1000_drive_data(radio,
			radio->air_transmit ? PIN_DRIVE_HIGH : PIN_DRIVE_LOW,
			PIN_DRIVE_LOW);
        radio->air_clockout = 0;
    } else {
        /* Rising edge */
        if(radio->sync) {
			
            /* Normal receive */
            if(!air_receive(radio->airwave, & radio->air_transmit,
				& radio->strength)) {
                /* Connection lost */
                radio->sync = 0;
                radio->sync_state = CC1000_SYNC_LOST;
            }
        } else {
            /* No signal */
			
            radio->air_transmit = 0;
        }
		
        cc1000_drive_data(radio, radio->pin_dio, PIN_DRIVE_HIGH);
        radio->air_clockout = 1;
    }
	
    return 0;
}

static int cc1000_clock_manreceive(void *device) {
    radio_cc1000 *radio = (radio_cc1000*)device;
    int bit, switched = 0;
    //
	
    /* Perform frequency locking and synchronization */
    if(!radio->sync) {
        if(((radio->sync_state == CC1000_SYNC_NOSIGNAL)
			&& !air_receive_changed(radio->airwave))
			|| !cc1000_receive_sync(radio)) {
			
            /* Not synchronized, just perform normal clock work */
            if(--radio->air_clock) {
                /* Between clock edges */
				
                return 0;
            }
			
            radio->air_clock = radio->air_clockoriginal;
            radio->air_clockdelta = radio->air_clockoriginal;
			
			
            radio->air_manchester = (radio->air_manchester + 1) & 0x03;
            cc1000_drive_data(radio, PIN_DRIVE_LOW,
				GET_FLAG(radio->air_manchester, 0x02)
				? PIN_DRIVE_HIGH : PIN_DRIVE_LOW);
            return 0;
        } else {
            /* Synchronization just completed, begin clocking */
            radio->air_clockdelta = radio->sync_cyclecount
                / radio->sync_bitcount / 4;
			
            /* During the preamble, all transitions occur at phase 2 */
            radio->air_manchester = 2;
            radio->air_clock = radio->air_clockdelta;
            radio->air_transmit = radio->sync_bit;
            radio->air_received = (radio->sync_bit
				? PIN_DRIVE_LOW : PIN_DRIVE_HIGH);
			
            /* Drive the clock and data initially */
            cc1000_drive_data(radio, PIN_DRIVE_LOW, PIN_DRIVE_HIGH);
        }
    }
	
    /* More complex than any other receive because we need to be constantly
	* adjusting the timing and receive through noise, etc.  Encoding errors
	* may force us to jump to different manchester phases and other nonsense.
	*/
	
    radio->air_clock--;
	//VMNet test...
	/* if(g_idChip)
    {
	if(tempdebugflag)
	{
	
	  tempdebugflag=0;
	  }
}*/      
	//
	
    if(air_receive_changed(radio->airwave)) {
        if(!air_receive(radio->airwave, &bit, &radio->strength)) {
			
            // Connection lost /
            radio->sync = 0;
            radio->sync_state = CC1000_SYNC_LOST;
            // If this occurs at a clock cycle, would wrap around and break /
            radio->air_clock++;
            return 0;
        }
		
        /* Transitions are more important than clock events */
		
		//if(bit != radio->air_transmit)  //whj remove 
        {
			
			//           radio->air_received = (bit ? PIN_DRIVE_HIGH:PIN_DRIVE_LOW );//whj
            switch(radio->air_manchester) {
            case 0:
                /* The transition was delayed, adjust timing to compensate */
                radio->air_clockdelta += (radio->air_clockdelta
					- radio->air_clock) / 2;
				
                radio->air_clock = radio->air_clockdelta;
				
                cc1000_drive_data(radio, radio->pin_dio, PIN_DRIVE_LOW);
                break;
            case 1:
                /* Transition came early, adjust timing to compensate */
                radio->air_clockdelta -= radio->air_clock / 2;
				
                radio->air_manchester = 2;
                radio->air_clock = radio->air_clockdelta;
                /* Clock transition at this stage */
                
				
                cc1000_drive_data(radio, radio->pin_dio, PIN_DRIVE_HIGH);
                /* Sample for driving later, if the current bit is low, was
				* a high to low transition, marking it as a 1.  Otherwise,
				* the opposite occured and it is a 0. */
                radio->air_received = (bit ?  PIN_DRIVE_LOW:PIN_DRIVE_HIGH);//whj
                break;
            case 2:
                /* Transition came very early, adjust timing to compensate */
                radio->air_clockdelta -= radio->air_clock;
                radio->air_clockdelta /= 2;
				
                radio->air_manchester = 0;
                radio->air_clock = radio->air_clockdelta;
				
				
                cc1000_drive_data(radio, radio->air_received, PIN_DRIVE_LOW);
				
                break;
            case 3:
                /* Transition came early, adjust timing to compensate */
                radio->air_clockdelta -= radio->air_clock / 2;
				
                radio->air_manchester = 0;
                radio->air_clock = radio->air_clockdelta;
                cc1000_drive_data(radio, radio->air_received, PIN_DRIVE_LOW);
				
                break;
				
            case -1:
            default:
			/* Last-ditch wait in phase 1, waiting for transition,
				* compensate and enter phase 2 */
                radio->air_clockdelta += (2 * radio->air_clockdelta
					- radio->air_clock) / 2;
                radio->air_manchester = 2;
                radio->air_clock = radio->air_clockdelta;
                /* Clock transition at this stage */
                cc1000_drive_data(radio, radio->pin_dio, PIN_DRIVE_HIGH);
                /* Sample for driving later, if the current bit is low, was
				* a high to low transition, marking it as a 1.  Otherwise,
				* the opposite occured and it is a 0. */
                radio->air_received = (bit ? PIN_DRIVE_LOW : PIN_DRIVE_HIGH);
                break;
            }
			
            switched = 1;
            radio->air_transmit = bit; 
        } /*if(bit != radio->air_transmit) {*/
    } /*if(air_receive_changed(radio->airwave)) {*/
	
    /* Check for clock ticks */
    if(radio->air_clock == 0) {
        /* Clock transition */
        radio->air_clock = radio->air_clockdelta;
        /* Clock transitions are only significant if no data bit transition */
        if(!switched) {
            switch(radio->air_manchester) {
            case 0:
                /* Transition to state 1 */
                radio->air_manchester = 1;
                break;
            case 1:
			/* We should had a transition, something is broken, go into
				* last-ditch wait (may be 180 degrees out of phase) */
                radio->air_manchester = -1;
                radio->air_clock = 2 * radio->air_clockdelta;
				LOG_CURRENT_FUNCTION("radiocc_cc1000_clock_manreceive",140);
                break;
            case 2:
                /* Transition to state 3 */
                radio->air_manchester = 3;
                break;
            case 3:
                /* Transition to state 0 */
                radio->air_manchester = 0;
                cc1000_drive_data(radio, radio->air_received, PIN_DRIVE_LOW);
                break;
            case -1:
            default:
                /* Last-ditch wait failed, carrier lost */
                radio->sync = 0;
                radio->sync_state = CC1000_SYNC_LOST;
                radio->air_clockdelta = radio->air_clockoriginal;
                radio->air_clock = radio->air_clockdelta;
                break;
            }
        } /*if(!switched) {*/
    } /*if(radio->air_clock == 0) {*/
	
    return 0;
}
