/*
* Copyright (c) 2003, University of Maryland 
* Copyright (c) 2003, Center for Satellite and Hybrid Communication Networks 
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions 
* are met:
*
*   * Redistributions of source code must retain the above copyright notice, 
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the University of Maryland/Center for Satellite
*     and Hybrid Communication Networks nor the names of its 
*     contributors may be used to endorse or promote products derived from 
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
* POSSIBILITY OF SUCH DAMAGE.
*
*/

/* Authors: 
*     Jonathan Polley (jpolley@isr.umd.edu),
*     Dionysus Blazakis (dblaze@isr.umd.edu),
*     Jonathan McGee (mcgee@isr.umd.edu),
*     Daniel Rusk (ruskd@isr.umd.edu),
*     Karl Seamon (kks@isr.umd.edu),
*     Manish Karir (karir@isr.umd.edu)
*/


/*
* Modifications:
*
* Copyright (c) 2004, The Hong Kong University of Science and Technology (HKUST)
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions
* are met:
*
*   * Redistributions of source code must retain the above copyright notice,
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the university nor the names of its
*     contributors may be used to endorse or promote products derived from
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
* POSSIBILITY OF SUCH DAMAGE.
*
*/


/*

  * Modification author: Hejun Wu (vmnet@cs.ust.hk)
  
********************************************************************************/

#include <assert.h>
#include "spi.h"
/*
* This module implements the Atmega128 SPI (Serial Peripheral Interface)
* device.  Currently only the slave mode is supported (master mode would
* not function well with the MICA2's radio).
*
* Registers:
*
* CONTROL - SPI Control Register, requires the entire byte
*    Bind to SPCR
* STATUS - SPI Status Register, requires the entire byte
*    Bind to SPSR
* DATA - SPI Data Register, requires the entire byte
*    Bind to SPDR
*
* DDRB - Data Direction Register for Port B
*    Bind to DDRB, mask 0x0F
* PORTB - Output Register for Port B
*    Bind to PORTB, mask 0x0F
* PINB - Input Register for Port B
*    Bind to PINB, mask 0x0F
*
*
* Interrupts:
*
* SPI - SPI Interrupt Request
*    Bind to 18 ($0022)
*
* Device also requires a CLOCK
*/

/* Local Static functions */
static void spi_destroy_device(void* device);
static error_t spi_enter_sleep(void* device, uint8_t sleepmode);
static error_t spi_leave_sleep(void *device);
static int spi_clock_master(void *device);
static int spi_clock_slave(void *device);
static int spi_clearint(void *device);
static uint8_t spi_control_read(void *dev);
static void spi_control_write(void *dev, uint8_t value);
static uint8_t spi_status_read(void *dev);
static void spi_status_write(void *dev, uint8_t value);
static uint8_t spi_data_read(void *dev);
static void spi_data_write(void *dev, uint8_t value);
static uint8_t spi_ddrb_read(void *dev);
static void spi_ddrb_write(void *dev, uint8_t value);
static uint8_t spi_pinb_read(void *dev);
static void spi_pinb_write(void *dev, uint8_t value);
static uint8_t spi_portb_read(void *dev);
static void spi_portb_write(void *dev, uint8_t value);

/*
* Startup functions
*/
device_obj* create_device(char *name) {
    spi_device *new_dev = (spi_device*)malloc(sizeof(spi_device));
    if(new_dev == NULL) {
        return NULL;
    }
	
    memset(new_dev, 0, sizeof(spi_device));
	
    new_dev->device.destroy = spi_destroy_device;
    new_dev->device.sleep = spi_enter_sleep;
    new_dev->device.wake = spi_leave_sleep;
    new_dev->device.dev = new_dev;
    new_dev->clk_tick.tick = NULL;
    new_dev->clk_tick.dev = new_dev;
    new_dev->int_spi.clear = spi_clearint;
    new_dev->int_spi.dev = new_dev;
	
    new_dev->reg_control.read = spi_control_read;
    new_dev->reg_control.write = spi_control_write;
    new_dev->reg_control.dev = new_dev;
    new_dev->reg_status.read = spi_status_read;
    new_dev->reg_status.write = spi_status_write;
    new_dev->reg_status.dev = new_dev;
    new_dev->reg_data.read = spi_data_read;
    new_dev->reg_data.write = spi_data_write;
    new_dev->reg_data.dev = new_dev;
	
    new_dev->reg_ddrb.read = spi_ddrb_read;
    new_dev->reg_ddrb.write = spi_ddrb_write;
    new_dev->reg_ddrb.dev = new_dev;
    new_dev->reg_portb.read = spi_portb_read;
    new_dev->reg_portb.write = spi_portb_write;
    new_dev->reg_portb.dev = new_dev;
    new_dev->reg_pinb.read = spi_pinb_read;
    new_dev->reg_pinb.write = spi_pinb_write;
    new_dev->reg_pinb.dev = new_dev;
	
    new_dev->spi_shifts = 8;
	
    /* External pins initally float */
    new_dev->pin_clock = new_dev->pin_clockin = PIN_FLOAT;
    new_dev->pin_miso = new_dev->pin_misoin = PIN_FLOAT;
    new_dev->pin_mosi = new_dev->pin_mosiin = PIN_FLOAT;
    new_dev->pin_slave = new_dev->pin_slavein = PIN_FLOAT;
    new_dev->pin_clock_prev = new_dev->pin_slave_prev = PIN_FLOAT;
	
    return &(new_dev->device);
}

register_obj *add_register(void *dev, char *name) {
    spi_device *spi = (spi_device*)dev;
	
    if(strcasecmp(name, "CONTROL") == 0) {
        return &(spi->reg_control);
    } else if(strcasecmp(name, "STATUS") == 0) {
        return &(spi->reg_status);
    } else if(strcasecmp(name, "DATA") == 0) {
        return &(spi->reg_data);
    } else if(strcasecmp(name, "DDRB") == 0) {
        return &(spi->reg_ddrb);
    } else if(strcasecmp(name, "PORTB") == 0) {
        return &(spi->reg_portb);
    } else if(strcasecmp(name, "PINB") == 0) {
        return &(spi->reg_pinb);
    }
    return NULL;
}

interrupt_obj* add_interrupt(void *dev, char *name) {
    spi_device *spi = (spi_device*)dev;
	
    if(strcasecmp(name, "SPI") == 0) {
        return &(spi->int_spi);
    }
	
    return NULL;
}

clock_obj* add_clock(void *dev) {
    spi_device *spi = (spi_device*)dev;
    return &(spi->clk_tick);
}

/*
* Basic device callbacks
*/

static void spi_destroy_device(void *dev) {
    spi_device *spi = (spi_device*)dev;
	
    free(spi);
}

static error_t spi_enter_sleep(void *dev, uint8_t sleepmode) {
    return 0;
}

static error_t spi_leave_sleep(void *dev) {
    return 0;
}

static int spi_clearint(void *dev) {
    spi_device *spi = (spi_device*)dev;
    spi->spi_intflag = 0;
	
    return 0;
}

static int spi_clock_master(void *dev) {
    /* Master mode not implemented, not functional with MICA2 radio */
	
    spi_device *spi = (spi_device*)dev;
    device_clrclock(spi->device, spi->clk_tick);
    return 0;
}

/*
* Functions to wrap pin changes for external hardware configuration
*/

static inline void spi_setpin_clock(spi_device* spi, pin_drive_enum value) {
    if(GET_FLAG(spi->portb_trueddr, SPI_PORTB_SCK)) {
        /* Can drive, see if it really has changed */
        if(value != spi->pin_clock) {
            spi->pin_clock = value; 
            if(spi->func_clock) {
                spi->func_clock(spi->func_clock_dev, spi, value);
            }
        }   
    }           
}           

static inline void spi_setpin_miso(spi_device* spi, pin_drive_enum value) {
    if(GET_FLAG(spi->portb_trueddr, SPI_PORTB_MISO)) {
        /* Can drive, see if it really has changed */
        if(value != spi->pin_miso) {
            spi->pin_miso = value;
            if(spi->func_miso) {
                spi->func_miso(spi->func_miso_dev, spi, value);
            }
        }
    }
}

static inline void spi_setpin_mosi(spi_device* spi, pin_drive_enum value) {
    if(GET_FLAG(spi->portb_trueddr, SPI_PORTB_MOSI)) {
        /* Can drive, see if it really has changed */
        if(value != spi->pin_mosi) {
            spi->pin_mosi = value;
            if(spi->func_mosi) {
                spi->func_mosi(spi->func_mosi_dev, spi, value);
            }
        }
    }
}

static inline void spi_setpin_slave(spi_device* spi, pin_drive_enum value) {
    if(GET_FLAG(spi->portb_trueddr, SPI_PORTB_SS)) {
        /* Can drive, see if it really has changed */
        if(value != spi->pin_slave) {
            spi->pin_slave = value;
            if(spi->func_slave) {
                spi->func_slave(spi->func_slave_dev, spi, value);
            }
        }
    }
}

/*
* SPI interface
*/

static uint8_t spi_control_read(void *dev) {
    spi_device *spi = (spi_device*)dev;
    return spi->spi_control;
}

static void spi_control_write(void *dev, uint8_t value) {
    spi_device *spi = (spi_device*)dev;
	
    /* Check if interrupt is signalled */
    spi->spi_intmask = GET_FLAG(value, SPI_SPCR_SPIE);
    if(spi->spi_intmask) {
        if(spi->spi_intflag) {
            spi->int_spi.set_interrupt(& spi->int_spi);
        }
    } else {
        spi->int_spi.clear_interrupt(& spi->int_spi);
    }
	
    /* See if there has been a change in the SPI state */
    if(GET_FLAG(value, SPI_SPCR_SPE)
		!= GET_FLAG(spi->spi_control, SPI_SPCR_SPE)) {
        /* Enabled or disabled? */
        if(GET_FLAG(value, SPI_SPCR_SPE)) {
            spi->device.device_debug(&spi->device, V_DEBUG,
				"SPI has been enabled.\n");
        } else {
            spi->device.device_debug(&spi->device, V_DEBUG,
				"SPI has been disabled.\n");
        }
        /* Call the DDR and PORT writes to fix the DDR and PINs */
        spi_portb_write(dev, spi->portb_out);
        spi_ddrb_write(dev, spi->portb_ddr);
    }
	
    /* Set clock appropriate to master slave */
    if(GET_FLAG(value, SPI_SPCR_SPE)) {
        if((value & SPI_SPCR_MSTR) == SPI_SPCR_MSTR_MASTER) {
            spi->device.device_debug(&spi->device, V_DEBUG,
				"SPI set to master.  This mode is NOT supported.\n");
            device_setclock(spi->device, spi->clk_tick, spi_clock_master);
        } else {
            spi->device.device_debug(&spi->device, V_DEBUG,
				"SPI set to slave.\n");
            spi->pin_slave_prev = PIN_FLOAT_HIGH;
			spi_setpin_mosi(spi, PIN_FLOAT);
        }
    }
	
    spi->spi_control = value;
}

static uint8_t spi_status_read(void *dev) {
    spi_device *spi = (spi_device*)dev;
    uint8_t value = 0;
	
    /* Construct return value */
    if(spi->spi_intflag) {
        SET_FLAG(value, SPI_SPSR_SPIF);
    }
    if(spi->spi_collision) {
        SET_FLAG(value, SPI_SPSR_WCOL);
        spi->spi_colread = 1;
    }
    if(spi->spi_spi2x) {
        SET_FLAG(value, SPI_SPSR_SPI2X);
    }
	
    return value;
}

static void spi_status_write(void *dev, uint8_t value) {
    spi_device *spi = (spi_device*)dev;
	
    /* See if interrupt is cleared */
    if(GET_FLAG(value, SPI_SPSR_SPIF)) {
        spi->spi_intflag = 0;
        spi->int_spi.clear_interrupt(& spi->int_spi);
    }
    /* Set the SPI2X flag */
    spi->spi_spi2x = GET_FLAG(value, SPI_SPSR_SPI2X);
}

static uint8_t spi_data_read(void *dev) {
    spi_device *spi = (spi_device*)dev;
	
    /* Clear collisions on read */
    if(spi->spi_colread) {
        spi->spi_collision = 0;
    }
    return spi->spi_buffer;
}
g_debugFlag0=0;
static void spi_data_write(void *dev, uint8_t value) {
    spi_device *spi = (spi_device*)dev;
	
    /* Check for collision */
    if((spi->spi_shifts != 0) && (spi->spi_shifts != 8)) {
        spi->spi_collision = 1;
        spi->spi_colread = 0;
    }
	
    /* Save output byte */
    spi->spi_register = value;
    /* If in master mode, begin a transfer */
    if((spi->spi_shifts == 0)
		&& ((spi->spi_control & SPI_SPCR_MSTR) == SPI_SPCR_MSTR_MASTER)) {
        spi->spi_shifts = 8;
    }
	
    spi->device.device_debug(&spi->device, V_DEBUG,
		"Byte 0x%02X queued for transmission on SPI.\n", value);
}

/*
* Port B interface
*/

static uint8_t spi_ddrb_read(void *dev) {
    spi_device *spi = (spi_device*)dev;
    return spi->portb_ddr;
}

static void spi_ddrb_write(void *dev, uint8_t value) {
    spi_device *spi = (spi_device*)dev;
    spi->portb_ddr = value;
	
    /* Set up the "real" DDR */
    if((spi->spi_control & SPI_SPCR_SPE) == 0) {
        /* SPI is not running, so a normal port */
        spi->portb_trueddr = value;
    } else if((spi->spi_control & SPI_SPCR_MSTR) == SPI_SPCR_MSTR_MASTER) {
        /* Running in master mode, MISO always an input */
        spi->portb_trueddr = (value & ~SPI_PORTB_MISO);
    } else {
        /* Running in slave mode, MOSI, SCK, and SS always inputs */
        spi->portb_trueddr = (value & SPI_PORTB_MISO);
    }
	
    /* Redrive the PORTD on DDR changes */
    spi_portb_write(dev, spi->portb_out);
}

static uint8_t spi_pinb_read(void *dev) {
    spi_device *spi = (spi_device*)dev;
    uint8_t value = 0;
	
    /* Slave Select */
    if(GET_FLAG(spi->portb_trueddr, SPI_PORTB_SS)) {
        if(GET_FLAG(spi->pin_slave, PIN_DRIVE_HIGH)) {
            SET_FLAG(value, SPI_PORTB_SS);
        }
    } else {
        if(GET_FLAG(spi->pin_slavein, PIN_DRIVE_HIGH)) {
            SET_FLAG(value, SPI_PORTB_SS);
        }
    }
	
    /* SPI Clock */
    if(GET_FLAG(spi->portb_trueddr, SPI_PORTB_SCK)) {
        if(GET_FLAG(spi->pin_clock, PIN_DRIVE_HIGH)) {
            SET_FLAG(value, SPI_PORTB_SCK);
        }
    } else {
        if(GET_FLAG(spi->pin_clockin, PIN_DRIVE_HIGH)) {
            SET_FLAG(value, SPI_PORTB_SCK);
        }
    }
	
    /* MISO (Master In - Slave Out) */
    if(GET_FLAG(spi->portb_trueddr, SPI_PORTB_MISO)) {
        if(GET_FLAG(spi->pin_miso, PIN_DRIVE_HIGH)) {
            SET_FLAG(value, SPI_PORTB_MISO);
        }
    } else {
        if(GET_FLAG(spi->pin_misoin, PIN_DRIVE_HIGH)) {
            SET_FLAG(value, SPI_PORTB_MISO);
        }
    }
	
    /* MOSI (Master Out - Slave In) */
    if(GET_FLAG(spi->portb_trueddr, SPI_PORTB_MOSI)) {
        if(GET_FLAG(spi->pin_mosi, PIN_DRIVE_HIGH)) {
            SET_FLAG(value, SPI_PORTB_MOSI);
        }
    } else {
        if(GET_FLAG(spi->pin_mosiin, PIN_DRIVE_HIGH)) {
            SET_FLAG(value, SPI_PORTB_MOSI);
        }
    }
	
    return value;
}

static void spi_pinb_write(void *dev, uint8_t value) {
    /* Writing to PINB does absolutely nothing */
}

static uint8_t spi_portb_read(void *dev) {
    spi_device *spi = (spi_device*)dev;
    return spi->portb_out;
}

static void spi_portb_write(void *dev, uint8_t value) {
    spi_device *spi = (spi_device*)dev;
    spi->portb_out = value;
	
    /* Copy over the bits onto the pins as appropriate for the modes */
    spi_setpin_slave(spi, GET_FLAG(value, SPI_PORTB_SS)
		? PIN_DRIVE_HIGH : PIN_DRIVE_LOW);
    if(!GET_FLAG(spi->spi_control, SPI_SPCR_SPE)) {
        /* We only have control of this port when the SPI is disabled */
        spi_setpin_clock(spi, GET_FLAG(value, SPI_PORTB_SCK)
			? PIN_DRIVE_HIGH : PIN_DRIVE_LOW);
        spi_setpin_mosi(spi, GET_FLAG(value, SPI_PORTB_MOSI)
			? PIN_DRIVE_HIGH : PIN_DRIVE_LOW);
        spi_setpin_miso(spi, GET_FLAG(value, SPI_PORTB_MISO)
			? PIN_DRIVE_HIGH : PIN_DRIVE_LOW);
    }
}

/*
* External device interface
*/

error_t spi_drive_slave(spi_device* device, pin_drive_enum value) {
    /* Check if slave is an input */
    if(GET_FLAG(device->portb_trueddr, SPI_PORTB_SS)) {
        if(GET_FLAG(value, PIN_FLOAT)) {
            /* External hardware trying to float when we are driving */
        } else {
            /* External hardware trying to drive when are driving */
            device->device.device_debug(&device->device, V_DEBUG,
				"External hardware driving SS when set as output.\n");
            return -1;
        }
    }
	
    /* See if it has changed */
    if(value != device->pin_slavein) {
        device->pin_slavein = value;
        /* Only set up the slave tick when the slave or clock occured */
        if((device->spi_control & SPI_SPCR_MSTR) == SPI_SPCR_MSTR_SLAVE) {
            device_setclock(device->device, device->clk_tick, spi_clock_slave);
        }
    }
	
    return 0;
}

error_t spi_drive_miso(spi_device* device, pin_drive_enum value) {
    /* Check if miso is an input */
    if(GET_FLAG(device->portb_trueddr, SPI_PORTB_MISO)) {
        if(GET_FLAG(value, PIN_FLOAT)) {
            /* External hardware trying to float when we are driving */
        } else {
            /* External hardware trying to drive when are driving */
            device->device.device_debug(&device->device, V_DEBUG,
				"External hardware driving MISO when set as output.\n");
			
            return -1;
        }
    }
	
    /* We only care about the value of this bit on clocks */
	
    device->pin_misoin = value;
    return 0;
}

error_t spi_drive_mosi(spi_device* device, pin_drive_enum value) {
    /* Check if mosi is an input */
    if(GET_FLAG(device->portb_trueddr, SPI_PORTB_MOSI)) {
        if(GET_FLAG(value, PIN_FLOAT)) {
            /* External hardware trying to float when we are driving */
        } else {
            /* External hardware trying to drive when are driving */
            device->device.device_debug(&device->device, V_DEBUG,
				"External hardware driving MOSI when set as output.\n");
            return -1;
        }
    }
	
    /* We only care about the value of this bit on clocks */
	
	//    device->spi_control |= SPI_SPCR_CPHA;
    device->pin_mosiin = value;
    return 0;
}

error_t spi_drive_clk(spi_device* device, pin_drive_enum value) {
    /* Check if clock is an input */
    if(GET_FLAG(device->portb_trueddr, SPI_PORTB_SCK)) {
        if(GET_FLAG(value, PIN_FLOAT)) {
            /* External hardware trying to float when we are driving */
        } else {
            /* External hardware trying to drive when are driving */
            device->device.device_debug(&device->device, V_DEBUG,
				"External hardware driving CLK when set as output.\n");
            return -1;
        }
    }
	
    /* See if it has changed */
    if(device->pin_clockin != value) {
        /* It has toggled, run on next clock if slave */
        device->pin_clockin = value;
        if((device->spi_control & SPI_SPCR_MSTR) == SPI_SPCR_MSTR_SLAVE) {
            device_setclock(device->device, device->clk_tick, spi_clock_slave);
        }
    }
	
    return 0;
}

error_t spi_set_slave(spi_device* device, pin_drive_callback func, void *dev) {
    /* Don't allow multiple registrations */
    if(device->func_slave != NULL) {
        return -1;
    }
	
    /* Register */
    device->func_slave = func;
    device->func_slave_dev = dev;
    return 0;
}

error_t spi_set_miso(spi_device* device, pin_drive_callback func, void *dev) {
    /* Don't allow multiple registrations */
    if(device->func_miso != NULL) {
        return -1;
    }
	
    /* Register */
    device->func_miso = func;
    device->func_miso_dev = dev;
    return 0;
}

error_t spi_set_mosi(spi_device* device, pin_drive_callback func, void *dev) {
    /* Don't allow multiple registrations */
    if(device->func_mosi != NULL) {
        return -1;
    }
	
    /* Register */
    device->func_mosi = func;
    device->func_mosi_dev = dev;
    return 0;
}

error_t spi_set_clk(spi_device* device, pin_drive_callback func, void *dev) {
    /* Don't allow multiple registrations */
    if(device->func_clock != NULL) {
        return -1;
    }
	
    /* Register */
    device->func_clock = func;
    device->func_clock_dev = dev;
    return 0;
}

/*
* Internal Device Functions
*/

static int spi_clock_slave(void *dev) {
    spi_device *spi = (spi_device*)dev;
    int leading;
    //whj debug
	
	
    /* In slave mode, we are only driven once clock or ss changes.  Remove
	* us from the clock list before anything may wish to re-add us. */
    device_clrclock(spi->device, spi->clk_tick);
	
    /* Check if we are selected */
    if(GET_FLAG(spi->pin_slavein, PIN_DRIVE_LOW)
		!= GET_FLAG(spi->pin_slave_prev, PIN_DRIVE_LOW)) {
        /* If going high, we reset the state */
        if(!GET_FLAG(spi->pin_slavein, PIN_DRIVE_LOW)) {
            spi->spi_register = 0;
        }
		
        /* Begin shifting of another eight bits */
        spi->spi_shifts = 8;
		
        /* Save the current state for the next cycle */
        spi->pin_slave_prev = spi->pin_slavein;
    }
    /* We are only active when low */
    if(!GET_FLAG(spi->pin_slavein, PIN_DRIVE_LOW)) {
        return 0;
    }
	
    /* See if the clock has changed */
    if(GET_FLAG(spi->pin_clock_prev, PIN_DRIVE_HIGH)
		== GET_FLAG(spi->pin_clockin, PIN_DRIVE_HIGH)) {
        return 0;
    } else {
        spi->pin_clock_prev = spi->pin_clockin;
    }
	
    /* See if this is the leading clock edge */
    if(GET_FLAG(spi->pin_clockin, PIN_DRIVE_HIGH)) {
        /* Rising clock edge */
        leading = ((spi->spi_control & SPI_SPCR_CPOL)
			== SPI_SPCR_CPOL_LEADRISE);
    } else {
        /* Falling clock edge */
        leading = ((spi->spi_control & SPI_SPCR_CPOL)
			== SPI_SPCR_CPOL_LEADFALL);
    }
	
    /* See if we are sampling the input or driving the output.  In order to
	* simplify life, all changes occur on the /trailing/ edge. This excludes
	* the shift count.  We change this whenever we output a bit so that we
	* can easily check it in data_write(). */
    if(leading) {
        /* We are on the leading edge of the clock */
		
        if((spi->spi_control & SPI_SPCR_CPHA) == SPI_SPCR_CPHA_LEADSETUP) {
            /* We set-up a bit on the leading edge */
            spi->spi_shifts--;
            if((spi->spi_control & SPI_SPCR_DORD) == SPI_SPCR_DORD_MSB) {
                /* We setup the MSB first */
                spi_setpin_miso(spi, GET_FLAG(spi->spi_register, 0x80)
					? PIN_DRIVE_HIGH : PIN_DRIVE_LOW);
            } else {
                /* We setup the LSB first */
                spi_setpin_miso(spi, GET_FLAG(spi->spi_register, 0x01)
					? PIN_DRIVE_HIGH : PIN_DRIVE_LOW);
            }
        } else {
		/* We sense a bit on the leading edge, buffer until we can
			* shift during the trailing edge. */
            spi->pin_lastread = GET_FLAG(spi->pin_mosiin, PIN_DRIVE_HIGH);
        }
    } else {
        /* We are on the trailing edge of the clock */
		
        if((spi->spi_control & SPI_SPCR_CPHA) == SPI_SPCR_CPHA_LEADSETUP) {
            //  spi->spi_control &= (~SPI_SPCR_CPHA);
			
            /* We sense a bit on the trailing edge */
            if((spi->spi_control & SPI_SPCR_DORD) == SPI_SPCR_DORD_MSB) {
                /* Store into the LSB after a left shift */
                spi->spi_register <<= 1;
                if(GET_FLAG(spi->pin_mosiin, PIN_DRIVE_HIGH)) {
                    SET_FLAG(spi->spi_register, 0x01);
                }
				
            } else {
                /* Store into the MSB after a right shift */
                spi->spi_register >>= 1;
                if(GET_FLAG(spi->pin_mosiin, PIN_DRIVE_HIGH)) {
                    SET_FLAG(spi->spi_register, 0x80);
                }
            }
        } else {
            /* We set-up a bit on the trailing edge */
            if(spi->spi_shifts==8)
            {
				//              if(spi->spi_register)
				
			}
            spi->spi_shifts--;
            if((spi->spi_control & SPI_SPCR_DORD) == SPI_SPCR_DORD_MSB) {
                /* We setup the MSB first */
                spi_setpin_miso(spi, GET_FLAG(spi->spi_register, 0x80)
					? PIN_DRIVE_HIGH : PIN_DRIVE_LOW);
				
                spi->spi_register <<= 1;
				
                if(spi->pin_lastread) { //
					//if(GET_FLAG(spi->pin_mosiin, PIN_DRIVE_HIGH)){ //whj changed to here
                    SET_FLAG(spi->spi_register, 0x01);
                }
            } else {
                /* We setup the LSB first */
                spi_setpin_miso(spi, GET_FLAG(spi->spi_register, 0x01)
					? PIN_DRIVE_HIGH : PIN_DRIVE_LOW);
				
                spi->spi_register >>= 1;
                if(spi->pin_lastread) {
                    SET_FLAG(spi->spi_register, 0x80);
					
					
                }
            }
        }
    }
    
    /* When the shift count reaches 0, a full byte has been sent/received */
    if(spi->spi_shifts == 0) {
        spi->spi_shifts = 8;
        spi->spi_intflag = 1;
        spi->spi_buffer = spi->spi_register;
		
        if(spi->spi_intmask) {
			//if(spi->spi_register)
			
			
            spi->int_spi.set_interrupt(& spi->int_spi);
        }
        spi->device.device_debug(&spi->device, V_DEBUG,
			"Read a byte 0x%02X from the SPI\n", spi->spi_buffer);
    }
	
    return 0;
}
