/*
* Copyright (c) 2003, University of Maryland 
* Copyright (c) 2003, Center for Satellite and Hybrid Communication Networks 
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions 
* are met:
*
*   * Redistributions of source code must retain the above copyright notice, 
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the University of Maryland/Center for Satellite
*     and Hybrid Communication Networks nor the names of its 
*     contributors may be used to endorse or promote products derived from 
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
* POSSIBILITY OF SUCH DAMAGE.
*
*/

/* Authors: 
*     Jonathan Polley (jpolley@isr.umd.edu),
*     Dionysus Blazakis (dblaze@isr.umd.edu),
*     Jonathan McGee (mcgee@isr.umd.edu),
*     Daniel Rusk (ruskd@isr.umd.edu),
*     Karl Seamon (kks@isr.umd.edu),
*     Manish Karir (karir@isr.umd.edu)
*/

#include "timer8.h"
#include "../EM/debug.h"
/*
* This module implements the basic behavior for the 8-bit timers (timer0
* and timer2).  16-bit operation will be in a seperate file.  The following
* registers are defined:
*
* CONTROL - Timer Control word, each timer requires the entire byte
*    Timer 0 binds to TCCR0, Timer 2 binds to TCCR2
* COMPARE - Timer Compare word, each timer requires the entire byte
*    Timer 0 binds to OCR0, Timer 2 binds to OCR2
* COUNT - Timer Count word, each timer requires the entire byte
*    Timer 0 binds to TCNT0, Timer 2 binds to TCNT2
* INTMASK - Timer Interrupt mask, each timer needs two bits from byte
*    Both timers bind to TIMSK, Timer 0 mask 0x03, Timer 2 mask 0xB0
* INTFLAG - Timer Interrupt flag, each timer needs two bits from byte
*    Both timers bind to TIFR, Timer 0 mask 0x03, Timer 2 mask 0xB0
* TSM - Single bit from Special Function IO Register
*    Both timers bind to SFIOR, mask 0x80
* PSR - Single bit from Special Function IO Register, different for each timer
*    Both timers bind to SFIOR, Timer 0 mask 0x02, Timer 2 mask 0x01
*
* AYNC - Asychronous Status word, only exists on timer 0, needs lower nibble
*    Timer 0 binds to ASSR, mask 0x0F
*
* DDR - Timer 2 uses a pin on Port D to allow software/external clocking
*    Timer 2 binds to DDRE, mask 0x80
* PORT - Timer 2 uses a pin on Port D to allow software/external clocking
*    Timer 2 binds to PORTE, mask 0x80
*
* Interrupts are defined:
*
* OVERFLOW - Triggered when a clock overflow occurs
*    Timer 0 binds to 17 ($0020), Timer 2 binds to 11 ($0014)
* COMPARE - Triggered when a clock comparision occurs
*    Timer 0 binds to 16 ($001E), Timer 2 binds to 10 ($0012)
*/

/* List of divisors for each prescale selection (timer 0) */
static const uint16_t g_prescalers0[] = { 0, 1, 8, 32, 64, 128, 256, 1024 }; //whj changed
//static const uint16_t g_prescalers0[] = { 0, 1, 8,  64,  256, 1024 };
/* List of divisors for each prescale selection (timer 2) */
static const uint16_t g_prescalers2[] = { 0, 1, 8, 64, 256, 1024, -1, -1 };

/* Local Static Functions */
static void timer8_destroy_device(void *device);
static error_t timer8_enter_sleep(void *device, uint8_t sleepmode);
static error_t timer8_leave_sleep(void *device);
static int timer8_clearint_compare(void *device);
static int timer8_clearint_overflow(void *device);
static uint8_t timer8_control_read(void *dev);
static void timer8_control_write(void *dev, uint8_t value);
static uint8_t timer8_compare_read(void *dev);
static void timer8_compare_write(void *dev, uint8_t value);
static uint8_t timer8_count_read(void *dev);
static void timer8_count_write(void *dev, uint8_t value);
static uint8_t timer8_async_read(void *dev);
static void timer8_async_write(void *dev, uint8_t value);
static uint8_t timer8_intflag_read(void *dev);
static void timer8_intflag_write(void *dev, uint8_t value);
static uint8_t timer8_intmask_read(void *dev);
static void timer8_intmask_write(void *dev, uint8_t value);
static uint8_t timer8_tsm_read(void *dev);
static void timer8_tsm_write(void *dev, uint8_t value);
static uint8_t timer8_psr_read(void *dev);
static void timer8_psr_write(void *dev, uint8_t value);
static int timer8_check_control(timer8_device *timer);
static int timer8_clock_async(void *device);
static int timer8_clock_cpu(void *device);
static int timer8_clock_external(void *device);
static int timer8_clock_common(timer8_device *timer);

/*
* Startup functions
*/
device_obj* create_device(char *name) {
	static sCounter=0;
    timer8_device *new_dev = (timer8_device*)malloc(sizeof(timer8_device));
    if(new_dev == NULL) {
        return NULL;
    }
	
    memset(new_dev, 0, sizeof(timer8_device));
    /* Default 32kHz clock for a 8MHz Atmega */
    new_dev->async_divisor = 250;
    DEBUG_TIMER("Created a timer, :::%d", sCounter++);
    /* Set to TIMER0 prescalers if ASYNC is requested */
    new_dev->prescalers = g_prescalers2;
	
    new_dev->device.destroy = timer8_destroy_device;
    new_dev->device.sleep = timer8_enter_sleep;
    new_dev->device.wake = timer8_leave_sleep;
    new_dev->device.dev = new_dev;
    new_dev->clk_tick.tick = NULL;
    new_dev->clk_tick.dev = new_dev;
    new_dev->int_compare.clear = timer8_clearint_compare;
    new_dev->int_compare.dev = new_dev;
    new_dev->int_overflow.clear = timer8_clearint_overflow;
    new_dev->int_overflow.dev = new_dev;
	
    new_dev->reg_control.read = timer8_control_read;
    new_dev->reg_control.write = timer8_control_write;
    new_dev->reg_control.dev = new_dev;
    new_dev->reg_compare.read = timer8_compare_read;
    new_dev->reg_compare.write = timer8_compare_write;
    new_dev->reg_compare.dev = new_dev;
    new_dev->reg_count.read = timer8_count_read;
    new_dev->reg_count.write = timer8_count_write;
    new_dev->reg_count.dev = new_dev;
    new_dev->reg_intmask.read = timer8_intmask_read;
    new_dev->reg_intmask.write = timer8_intmask_write;
    new_dev->reg_intmask.dev = new_dev;
    new_dev->reg_intflag.read = timer8_intflag_read;
    new_dev->reg_intflag.write = timer8_intflag_write;
    new_dev->reg_intflag.dev = new_dev;
    new_dev->bit_tsm.read = timer8_tsm_read;
    new_dev->bit_tsm.write = timer8_tsm_write;
    new_dev->bit_tsm.dev = new_dev;
    new_dev->bit_psr.read = timer8_psr_read;
    new_dev->bit_psr.write = timer8_psr_write;
    new_dev->bit_psr.dev = new_dev;
	
    new_dev->reg_async.read = timer8_async_read;
    new_dev->reg_async.write = timer8_async_write;
    new_dev->reg_async.dev = new_dev;
	
    return &(new_dev->device);
}

register_obj *add_register(void *dev, char *name) {
    timer8_device *timer = (timer8_device*)dev;
	
    if(strcasecmp(name, "CONTROL") == 0) {
        return &(timer->reg_control);
    } else if(strcasecmp( name, "COMPARE" ) == 0) {
        return &(timer->reg_compare);
    } else if(strcasecmp( name, "COUNT" ) == 0) {
        return &(timer->reg_count);
    } else if(strcasecmp( name, "ASYNC" ) == 0) {
        timer->prescalers = g_prescalers0;
        return &(timer->reg_async);
    } else if(strcasecmp( name, "INTMASK" ) == 0) {
        return &(timer->reg_intmask);
    } else if(strcasecmp( name, "INTFLAG" ) == 0) {
        return &(timer->reg_intflag);
    } else if(strcasecmp( name, "TSM" ) == 0) {
        return &(timer->bit_tsm);
    } else if(strcasecmp( name, "PSR" ) == 0) {
        return &(timer->bit_psr);
    }
	
    return NULL;
}

interrupt_obj* add_interrupt(void *dev, char *name) {
    timer8_device *timer = (timer8_device*)dev;
	
    if(strcasecmp(name, "COMPARE") == 0) {
        return &(timer->int_compare);
    } else if(strcasecmp(name, "OVERFLOW") == 0) {
        return &(timer->int_overflow);
    }
	
    return NULL;
}

clock_obj* add_clock(void *dev) {
    timer8_device *timer = (timer8_device*)dev;
    return &(timer->clk_tick);
}

error_t set_param(void *dev, char *key, char *value) {
    timer8_device *timer = (timer8_device*)dev;
    uint16_t parse;
    char *endparse;
	
    if(strcasecmp(key, "ASYNC_DIVIDE") == 0) {
        parse = strtoul(value, &endparse, 0);
        if(endparse && *endparse) {
            /* Non-numeric character */
            return -1;
        }
        timer->async_divisor = parse;
        return 0;
    }
	
    return -1;
}

/*
* Basic device callbacks
*/

static void timer8_destroy_device(void *dev) {
    timer8_device *timer = (timer8_device*)dev;
	
    free(timer);
}

static error_t timer8_enter_sleep(void *dev, uint8_t sleepmode) {
    timer8_device *timer = (timer8_device*)dev;
	
    switch(sleepmode) {
    case SLEEPMODE_IDLE:
    case SLEEPMODE_ADCNOISE:
    case SLEEPMODE_EXTSTANDBY:
        /* Modes all timers operate in */
        timer->sleeping = 0;
        timer->asyncsleeping = 0;
        break;
    case SLEEPMODE_POWERDOWN:
        /* Modes all timers are sleeping */
        timer->sleeping = 1;
        timer->asyncsleeping = 1;
        timer->async_divide = 0;
        break;
    case SLEEPMODE_POWERSAVE:
        /* Modes an asynchronous Timer0 can run in */
        timer->sleeping = 1;
        timer->asyncsleeping = 0;
        break;
    case SLEEPMODE_STANDBY:
        /* Modes non-asynchronous timers can run in */
        timer->sleeping = 0;
        timer->asyncsleeping = 1;
        timer->async_divide = 0;
        break;
    default:
        /* Shouldn't be in this mode.... */
        timer->device.device_debug(&timer->device, V_NORMAL,
			"Unknown sleep mode %d!\n", sleepmode);
        timer->sleeping = 1;
        timer->asyncsleeping = 1;
        device_clrclock(timer->device, timer->clk_tick);
        return -1;
    }
	
    /* Disable clock as appropriate */
    if(timer->asyncsleeping && timer->async) {
        device_clrclock(timer->device, timer->clk_tick);
    }
    if(timer->sleeping && !timer->async) {
        device_clrclock(timer->device, timer->clk_tick);
    }
	
    return 0;
}

static error_t timer8_leave_sleep(void *dev) {
    timer8_device *timer = (timer8_device*)dev;
    timer->sleeping = 0;
    timer->asyncsleeping = 0;
    /* Reenable any normal modes */
    if((timer->control & TIMER8_CONTROL_CS0) != 0) {
        if(timer->async) {
            device_setclock(timer->device, timer->clk_tick, timer8_clock_async);
        } else if(!timer->externalclock) {
            device_setclock(timer->device, timer->clk_tick, timer8_clock_cpu);
        }
    }
    return 0;
}

static int timer8_clearint_compare(void *dev) {
    timer8_device *timer = (timer8_device*)dev;
    timer->signalcompare = 0;
    return 0;
}

static int timer8_clearint_overflow(void *dev) {
    timer8_device *timer = (timer8_device*)dev;
    timer->signaloverflow = 0;
    return 0;
}

/*
* Device I/O
*/

static uint8_t timer8_count_read(void *dev) {
    timer8_device *timer = (timer8_device*)dev;
    /* Count read always gives the current count, even if a previous write
	* is buffering a different count. */
	
	
    return timer->count;
}

static void timer8_count_write(void *dev, uint8_t value) {
    timer8_device *timer = (timer8_device*)dev;
	
	
    if(timer->async) {
        /* In asynchronous mode, count writes are buffered */
		
        timer->count_asyncbuffer = value;
        timer->asyncbufferedcount = 1;
    } else {
	/* According to the spec, writing to count will skip the comparision
		* of the next timer tick. */
        timer->skipnextcompare = 1;
        timer->bufferedcount = 1;
        timer->count = value;
    }
}

static uint8_t timer8_compare_read(void *dev) {
    timer8_device *timer = (timer8_device*)dev;
	
    if(timer->async) {
		
        return timer->compare_asyncbuffer;
    } else if(timer->buffercompare) {
		
        return timer->compare_buffer;
    } else {
		
        return timer->compare;
    }
}

static void timer8_compare_write(void *dev, uint8_t value) {
    timer8_device *timer = (timer8_device*)dev;
	
    if(timer->async) {
        timer->compare_asyncbuffer = value;
    } else if(timer->buffercompare) {
        timer->compare_buffer = value;
    } else {
        timer->compare = value;
    }
}

static uint8_t timer8_control_read(void *dev) {
    timer8_device *timer = (timer8_device*)dev;
    if(timer->async) {
        return timer->control_asyncbuffer;
    } else {
		
        return timer->control;
    }
}

static void timer8_control_write(void *dev, uint8_t value) {
    timer8_device *timer = (timer8_device*)dev;
	
    /* The most significant bit is write-only and outside our scope */
    value &= ~TIMER8_CONTROL_FOC0;
	
    if(timer->async) {
        timer->control_asyncbuffer = value;
		
        /* See if clock is being enabled */
        if((value & TIMER8_CONTROL_CS0) != 0) {
            device_setclock(timer->device, timer->clk_tick, timer8_clock_async);
        }
    } else {
        /* Copy and perform basic tests */
        timer->control = value;
		
        timer8_check_control(timer);
    }
}

static uint8_t timer8_async_read(void *dev) {
    timer8_device *timer = (timer8_device*)dev;
    uint8_t value = ((timer->async) ? TIMER8_ASSR_AS0 : 0);
	
    if(timer->async) {
        /* In asynchronous mode, test other registers */
        if(timer->bufferedcount || timer->asyncbufferedcount) {
            value |= TIMER8_ASSR_TCN0UB;
        }
        if(timer->compare_asyncbuffer != timer->compare_buffer) {
            value |= TIMER8_ASSR_OCR0UB;
        }
        if(timer->control != timer->control_asyncbuffer) {
            value |= TIMER8_ASSR_TCR0UB;
        }
    }
	
	
    return value;
}

static void timer8_async_write(void *dev, uint8_t value) {
    timer8_device *timer = (timer8_device*)dev;
	
    if(GET_FLAG(value, TIMER8_ASSR_AS0)) {
        /* Entering asynchronous mode */
        timer->async = 1;
        device_setclock(timer->device, timer->clk_tick, timer8_clock_async);
    } else {
        /* Leaving asynchronous mode */
        timer->async = 0;
        device_setclock(timer->device, timer->clk_tick, timer8_clock_cpu);
    }
}

static uint8_t timer8_intflag_read(void *dev) {
    timer8_device *timer = (timer8_device*)dev;
    return (timer->signalcompare ? TIMER8_TIFR_OCF0 : 0)
        | (timer->signaloverflow ? TIMER8_TIFR_TOV0 : 0);
}

static void timer8_intflag_write(void *dev, uint8_t value) {
    timer8_device *timer = (timer8_device*)dev;
    if(GET_FLAG(value, TIMER8_TIFR_OCF0)) {
        /* Clear compare interrupt */
        timer->signalcompare = 0;
    }
    if(GET_FLAG(value, TIMER8_TIFR_TOV0)) {
        /* Clear overflow interrupt */
        timer->signaloverflow = 0;
    }
}

static uint8_t timer8_intmask_read(void *dev) {
    timer8_device *timer = (timer8_device*)dev;
    return (timer->maskcompare ? TIMER8_TIMSK_OCIE0 : 0)
		| (timer->maskoverflow ? TIMER8_TIMSK_TOIE0 : 0);
}

static void timer8_intmask_write(void *dev, uint8_t value) {
    timer8_device *timer = (timer8_device*)dev;
    timer->maskcompare = GET_FLAG(value, TIMER8_TIMSK_OCIE0);
    timer->maskoverflow = GET_FLAG(value, TIMER8_TIMSK_TOIE0);
    /* Make sure the active interrupt table is up to date */
    if(timer->maskcompare && timer->signalcompare) {
        timer->int_compare.set_interrupt(& timer->int_compare);
    } else {
        timer->int_compare.clear_interrupt(& timer->int_compare);
    }
    if(timer->maskoverflow && timer->signaloverflow) {
        timer->int_overflow.set_interrupt(& timer->int_overflow);
    } else {
        timer->int_overflow.clear_interrupt(& timer->int_overflow);
    }
}

static uint8_t timer8_tsm_read(void *dev) {
    timer8_device *timer = (timer8_device*)dev;
    return (timer->enabletsm) ? 1 : 0;
}

static void timer8_tsm_write(void *dev, uint8_t value) {
    timer8_device *timer = (timer8_device*)dev;
    /* If TSM is cleared, all clocks restart */
    if(timer->enabletsm && !value) {
        timer->enablepsr = 0;
    }
    timer->enabletsm = value;
}

static uint8_t timer8_psr_read(void *dev) {
    timer8_device *timer = (timer8_device*)dev;
    return (timer->enablepsr) ? 1 : 0;
}

static void timer8_psr_write(void *dev, uint8_t value) {
    timer8_device *timer = (timer8_device*)dev;
    timer->enablepsr = GET_FLAG(value, 0x01);
    if(timer->enablepsr) {
        /* Prescaler held in reset, so reset running count */
        timer->clock_divide = 0;
    }
}

/*
* Main hardware logic
*/

static int timer8_check_control(timer8_device *timer) {
    /* Compares are buffered in the PWM modes */
    timer->buffercompare = (timer->control & TIMER8_CONTROL_WGM0_PWM);
    /* Select the prescaler */
    timer->prescaler_divisor
        = timer->prescalers[timer->control & TIMER8_CONTROL_CS0];
    /* Check if timer being disabled or disabled */
    if(!timer->async) {
        if(timer->prescaler_divisor == 0) {
            /* Clock disabled */
            device_clrclock(timer->device, timer->clk_tick);
            timer->externalclock = 0;
        } else if(timer->prescaler_divisor != (uint16_t)-1) {
            /* Clocked by CPU clock */
            device_setclock(timer->device, timer->clk_tick, timer8_clock_cpu);
            timer->externalclock = 0;
        } else {
            /* External clock generation */
            device_clrclock(timer->device, timer->clk_tick);
            timer->externalclock = 1;
        }
    }
    return 0;
}

static int timer8_clock_async(void *device) {
    timer8_device *timer = (timer8_device*)device;
    uint16_t prescalebit;
	
    /* Increment the asynchronous timer */
    timer->async_divide++;
    if(timer->async_divide == timer->async_divisor) {
        timer->async_divide = 0;
		
        /* In asynchronous mode, values written to the count, compare, and
		* control registers are latched into the real count, compare, etc.
		* after two rising transitions of the asynchronous clock.  We move
		* values from the triple buffer to the double buffer on each tick,
		* and the values are then transfered into the real buffers on the
		* next. */
		
        /* Move CONTROL byte from double buffer into active */
        if(timer->control != timer->control_buffer) {
            timer->control = timer->control_buffer;
            timer8_check_control(timer);
        }
		
        /* Compare may not be immediately copied in some modes */
        if(!timer->buffercompare) {
            timer->compare = timer->compare_buffer;
        }
		
        if(timer->bufferedcount) {
		/* If a count has been buffered, copy the new value and
			* don't perform any comparisions until after the next tick */
            timer->count = timer->count_buffer;
            timer->skipnextcompare = 1;
            timer->bufferedcount = 0;
        }
		
        /* Move compare and control forward one tick */
        timer->compare_buffer = timer->compare_asyncbuffer;
        timer->control_buffer = timer->control_asyncbuffer;
        /* Move count forward one buffer if necessary */
        if(timer->asyncbufferedcount) {
            timer->count_buffer = timer->count_asyncbuffer;
            timer->bufferedcount = 1;
            timer->asyncbufferedcount = 0;
        }
		
        /* Run through the prescaler */
        if(timer->prescaler_divisor == 0) {
            /* Clock is disabled */
        } else {
            /* Running through the prescaler */
            if(!timer->enablepsr) {
                /* If Prescale Reset is disabled, increment */
                timer->clock_divide++;
            } else if(!timer->enabletsm) {
                /* If PSR is set without TSM, PSR is cleared after one tick */
                timer->enablepsr = 0;
            }
            /* Take a look at the prescale bit */
            prescalebit = GET_FLAG(timer->clock_divide,
				timer->prescaler_divisor);
				/* The logic sees a tick occurs if the bit has toggled since the
            * last tick */
			
            if(prescalebit != timer->lasttick) {
				
                timer8_clock_common(timer);
            }
			
            timer->lasttick = prescalebit;
        }
    } /* if(timer->async_divide == timer->async_divisor) { */
	
    return 0;
}

static int timer8_clock_cpu(void *device) {
    timer8_device *timer = (timer8_device*)device;
    uint8_t tick = 0, prescaler, prescalebit;
	
    /* Run through the prescaler */
    prescaler = timer->prescalers[timer->control & TIMER8_CONTROL_CS0];
    if(prescaler == 0) {
        /* The clock is disabled, why are we here? */
        device_clrclock(timer->device, timer->clk_tick);
        tick = 0;
    } else {
        /* Running through the prescaler */
        if(!timer->enablepsr) {
            /* If Prescale Reset is disabled, increment */
            timer->clock_divide++;
        } else if(!timer->enabletsm) {
            /* If PSR is set without TSM, PSR is cleared after one tick */
            timer->enablepsr = 0;
        }
        /* Use the list for timer 0 */
        prescalebit = GET_FLAG(timer->clock_divide, prescaler);
        /* The logic sees a tick occurs if the bit has toggled since the
		* last tick */
        if(prescalebit != timer->lasttick) {
            timer8_clock_common(timer);
        }
        timer->lasttick = prescalebit;
    }
	
    return 0;
}

static int timer8_clock_external(void *device) {
    timer8_device *timer = (timer8_device*)device;
	
    /* Only called on change */
    device_clrclock(timer->device, timer->clk_tick);
	
    /* Make sure there has been a change */
    if(timer->externalvalue == timer->externalprevious) {
        return 0;
    }
	
    /* Make sure it is the correct transition */
    timer->externalprevious = timer->externalvalue;
    if(timer->externalvalue) {
        /* Rising edge */
        if(GET_FLAG(timer->control, TIMER8_CONTROL_CS00)) {
            /* We clock on rising edge */
            timer8_clock_common(timer);
        }
    } else {
        /* Falling edge */
        if(!GET_FLAG(timer->control, TIMER8_CONTROL_CS00)) {
            /* We clock on falling edge */
            timer8_clock_common(timer);
        }
    }
	
    return 0;
}

error_t timer8_drive_clock(timer8_device* timer, pin_drive_enum value) {
    unsigned bit = GET_FLAG(value, PIN_DRIVE_HIGH);
    /* For efficiency, only perform above function on clock change */
    if(bit != timer->externalvalue) {
        timer->externalvalue = bit;
        /* Don't touch clk_tick if not in an externally clocked mode */
        if(timer->externalclock) {
            device_setclock(timer->device, timer->clk_tick,
				timer8_clock_external);
        }
    }
	
    return 0;
}

static int timer8_clock_common(timer8_device *timer) {
    uint8_t intcomp = 0, intover = 0;
	
    /* Comparision made continuously */
    intcomp = !timer->skipnextcompare && (timer->count == timer->compare);
    /* Compares are skipped after a new count is written */
    timer->skipnextcompare = 0;
	
    /* Control Logic (overflow handling is specific to the waveform
	* generation mode, and is performed before increment.  It is supposed
	* to be timed for the tick /after/ the event actually occured) */
    switch(timer->control & TIMER8_CONTROL_WGM0) {
    case TIMER8_CONTROL_WGM0_MODE0:
        /* Normal Mode */
        intover = (timer->count == TIMER8_MAX);
        timer->count++;
        break;
    case TIMER8_CONTROL_WGM0_MODE2:
        /* Clear Timer on Compare Match Mode */
        if(intcomp) {
            timer->count = TIMER8_BOTTOM;
        } else {
            timer->count++;
        }
        break;
    case TIMER8_CONTROL_WGM0_MODE3:
        /* Fast PWM Mode */
        if(timer->count == TIMER8_MAX) {
            intover++;
            /* Compare is buffered until TOP */
            timer->compare = timer->compare_buffer;
        }
        timer->count++;
        break;
    case TIMER8_CONTROL_WGM0_MODE1:
        /* Phase Correct PWM Mode */
        if(timer->count == TIMER8_MAX) {
            /* Compare is buffered until TOP */
            timer->compare = timer->compare_buffer;
            timer->direction = 1;
        } else if(timer->count == TIMER8_BOTTOM) {
            /* Unlike other modes, PC PWM overflows on bottom */
            intover = 1;
            timer->direction = 0;
        }
        if(timer->direction) {
            timer->count--;
        } else {
            timer->count++;
        }
        break;
    }
	
    timer->device.device_debug(&timer->device, V_ALL,
		"count=%d, compare=%d, control=%02x\n",
		timer->count, timer->compare, timer->control);
	
    /* Handle new interrupts */
    if(intover) {
        if(!timer->lastoverflow) {
            timer->lastoverflow = 1;
            timer->signaloverflow = 1;
            if(timer->maskoverflow) {
				
                timer->int_overflow.set_interrupt(& timer->int_overflow);
            }
        }
    } else {
        timer->lastoverflow = 0;
    }
    if(intcomp) {
        if(!timer->lastcompare) {
            timer->lastcompare = 1;
            timer->signalcompare = 1;
            if(timer->maskcompare) {
                timer->int_compare.set_interrupt(& timer->int_compare);
            }
        }
    } else {
        timer->lastcompare = 0;
    }
	
    return 0;
}
