/*
* Copyright (c) 2004, The Hong Kong University of Science and Technology (HKUST)
* All rights reserved.
* 
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions
* are met:
*
*   * Redistributions of source code must retain the above copyright notice,
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the university nor the names of its
*     contributors may be used to endorse or promote products derived from
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
* POSSIBILITY OF SUCH DAMAGE.
*
*/


/*

 *  Author: Hejun Wu (vmnet@cs.ust.hk)
 
********************************************************************************/

//------------------------------------------------------------------------------
// bridge.c: Virtual UART module. Connect the virtual sink node with the PC that
//           runs the user client. The user client should provide an interface to
//           connect with this module.
// Created date: 2004-7-10
// Modified date: 2004-10-5
//------------------------------------------------------------------------------
#include "bridge.h"
#include "../EM/atmega128.h"
#include <pthread.h>
#include "../devices/usart.h"

#if defined(__CYGWIN__)
// for 64-bit byte-swap in Cygwin.
#define htonll(x) bswap_64(x)
#define ntohll(x) bswap_64(x)
#define swap_type(type,a,b) { type t=(a); (a)=(b); (b)=t; }

int64_t bswap_64( int64_t n )
{
	int8_t* p = (int8_t*)&n;
	swap_type( int8_t, p[0], p[7] );
	swap_type( int8_t, p[1], p[6] );
	swap_type( int8_t, p[2], p[5] );
	swap_type( int8_t, p[3], p[4] );
	return n;
}
#if __BYTE_ORDER == __BIG_ENDIAN
#  define htonll(x) (x)
#  define ntohll(x) (x)
#else
#  if __BYTE_ORDER == __LITTLE_ENDIAN
#    define htonll(x) __bswap_64(x)
#    define ntohll(x) __bswap_64(x)
#  endif
#endif

#endif//if defined(__CYGWIN__)

#define PACKET_SIZE 256

unsigned char UDR[PACKET_SIZE];

//These are only for test
#ifdef VUART_TEST
unsigned char testCommandMsg[3][100] = {
	{ 0x7E, 0x41, 0x0E, 0xFF, 0xFF, 0x6A, 0x7D, 0x5D, 0x0A, 0x01, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFE, 0xFD, 0x7E,},
	{ 0x7E, 0x41, 0x0F, 0xFF, 0xFF, 0x65, 0x7D, 0x5D, 0x28, 0x00, 0x7D, 0x5E, 0x00,
	0x00, 0x02, 0x00, 0xFF, 0x00, 0xCC, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x6E, 0x6F, 0x64, 0x65, 0x69, 0x64, 0x00, 0x00, 0x00, 0x04,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x9E, 0x30,
	0x7E,},
	{ 0x7E, 0x41, 0x10, 0xFF, 0xFF, 0x65, 0x7D, 0x5D, 0x28, 0x00, 0x7D, 0x5E, 0x00,
	0x00, 0x02, 0x00, 0xFF, 0x00, 0xCC, 0x00, 0x00, 0x01, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x6C, 0x69, 0x67, 0x68, 0x74, 0x00, 0x00, 0x00, 0x00, 0x04,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x48, 0x96,
	0x7E,},
};
#endif //#ifdef VUART_TEST

int commandServerSocket = -1;
int eventServerSocket = -1;
int commandClient;
int eventClient;

pthread_t clientAcceptThread;
pthread_t commandReadThread;
pthread_mutex_t commandClientsLock;
pthread_cond_t commandClientsCond;
pthread_mutex_t eventClientsLock;
pthread_cond_t eventClientsCond;


int createServerSocket(short port);
void *clientAcceptThreadFunc(void *arg);
void *commandReadThreadFunc(void *arg);

extern usart_device* g_usart=NULL;

void set_uart_int_rx(usart_device *usart) {
	usart->controlA |= USART_CONTROLA_RXC0;
	if(GET_FLAG(usart->controlB, USART_CONTROLB_RXCIE0)) {
        usart->int_rx.set_interrupt(&(usart->int_rx));
    }
}

//------------------------------------------------------------------------------
//  Initialization of sockets
//------------------------------------------------------------------------------

void initializeSockets() {
	int i;
	
	commandClient = -1;
	eventClient = -1;
	commandServerSocket = createServerSocket(COMMAND_PORT);
	eventServerSocket = createServerSocket(EVENT_PORT);
	pthread_mutex_init(&eventClientsLock, NULL);
	pthread_mutex_init(&commandClientsLock, NULL);
	pthread_cond_init(&commandClientsCond, NULL);
	pthread_cond_init(&eventClientsCond, NULL);
	pthread_create(&clientAcceptThread, NULL, clientAcceptThreadFunc, NULL);
	pthread_create(&commandReadThread, NULL, commandReadThreadFunc, NULL);
	socketsInitialized = 1;
}

//------------------------------------------------------------------------------
//  Socket management
//------------------------------------------------------------------------------

int acceptConnection(int servfd) {
	struct sockaddr_in cli_addr;
	int clilen = sizeof(cli_addr);
	int clifd;
	DEBUG_OUTPUT("Waiting for connection on socket %d .... \n", servfd);
	clifd = accept(servfd, (struct sockaddr*)&cli_addr, &clilen);
	if (clifd < 0) {
		DEBUG_OUTPUT("Could not accept socket: %s,  %d\n", strerror(errno),clifd);
		exit(-1);
	}
	return clifd;
}

int createServerSocket(short port) {
	struct sockaddr_in sock;
	int sfd;
	int rval = -1;
	long enable = 1;
	
	memset(&sock, 0, sizeof(sock));
	sock.sin_family = AF_INET;
	sock.sin_port = htons(port);
	sock.sin_addr.s_addr = htonl(INADDR_LOOPBACK);
	
	sfd = socket(AF_INET, SOCK_STREAM, 0);
	if (sfd < 0) {
		DEBUG_OUTPUT("Could not create server socket: %s\n", strerror(errno));
		exit(-1);
	}
	setsockopt(sfd, SOL_SOCKET, SO_REUSEADDR, (char *)&enable, sizeof(int));
	
	while(rval < 0) {
		rval = bind(sfd, (struct sockaddr*)&sock, sizeof(sock));
		if (rval < 0) {
			DEBUG_OUTPUT("Could not bind server socket to port %s\n", strerror(errno));
			sleep(10);
		}
	}
	
	if (listen(sfd, 1) < 0) {
		DEBUG_OUTPUT(" Could not listen on server socket: %s\n", strerror(errno));
		exit(-1);
	}
	DEBUG_OUTPUT("Created server socket listening on port %d.\n", port);
	return sfd;
}

void waitForConnection() {
	if (eventClient == -1) {
		pthread_cond_wait(&eventClientsCond, &eventClientsLock);
    }
}

//------------------------------------------------------------------------------
// Client accept thread
//------------------------------------------------------------------------------

void *clientAcceptThreadFunc(void *arg) {
	int clifd;
	fd_set acceptset;
	int highest = (commandServerSocket>eventServerSocket)?commandServerSocket:eventServerSocket;
	DEBUG_OUTPUT("clientAcceptThread running.\n");
	while (1) {
		usleep(100);    // make it not so frequently
		FD_ZERO(&acceptset);
		FD_SET(commandServerSocket, &acceptset);
		FD_SET(eventServerSocket, &acceptset);
		DEBUG_OUTPUT("clientAcceptThread: calling select...\n");
		if (select(highest+1, &acceptset, NULL, NULL, NULL) < 0) {
			DEBUG_OUTPUT("clientAcceptThreadFunc: error in select(): %s\n", strerror(errno));
		}
		/* Command socket */
		if (FD_ISSET(commandServerSocket, &acceptset)) {
			DEBUG_OUTPUT("clientAcceptThread: Checking for command connection\n");
			clifd = acceptConnection(commandServerSocket);
			DEBUG_OUTPUT(" clientAcceptThread: Got command connection %d\n", clifd);
			pthread_mutex_lock(&commandClientsLock);
			commandClient= clifd;
			pthread_cond_broadcast(&commandClientsCond);
			pthread_mutex_unlock(&commandClientsLock);
		}
		
		/* Event socket */
		if (FD_ISSET(eventServerSocket, &acceptset)) {
			DEBUG_OUTPUT("clientAcceptThread: Checking for event connection\n");
			clifd = acceptConnection(eventServerSocket);
			DEBUG_OUTPUT("clientAcceptThread: Got event connection %d\n", clifd);
			pthread_mutex_lock(&eventClientsLock);
			eventClient = clifd;
			pthread_cond_broadcast(&eventClientsCond);
			pthread_mutex_unlock(&eventClientsLock);
		}
	}
	return 0;
}

//------------------------------------------------------------------------------
//  Writing BYTES TO BASE STATION.
//------------------------------------------------------------------------------

void transmitEvent(int clifd, uint16_t type, void *data, int dataLen) {
	if (!socketsInitialized) return;
	char * msg;
	msg=(char*)malloc(dataLen+2);
	if(!msg)
		return;
	msg[0] = (dataLen&0xff00)>>8;
	msg[1] = (dataLen&0xff);
	memcpy(msg+2,data,dataLen);
	// If no clients and '-gui', wait for a connection
	pthread_mutex_lock(&eventClientsLock);
	if (eventClient==-1)
	{
		pthread_mutex_unlock(&eventClientsLock);
		waitForConnection();    
	}
	if (send(eventClient, msg, dataLen+2, 0) < 0) {
		eventClient = -1;
		DEBUG_OUTPUT(" NO CONNECTION ERR: %d\n", type, dataLen);
	}
    pthread_mutex_unlock(&eventClientsLock);
    DEBUG_OUTPUT(" NO  ERR: %d\n",  dataLen);
}

//-----------------------------------------------------------------------------------------
// Read in a command from the base station.
// Returns 0 if successful, -1 if the client connection was closed
//------------------------------------------------------------------------------------------

int recvCommand(int clifd, void* data) {
	int curlen = 0;
	int rval;
	
	DEBUG_OUTPUT("Reading command from client fd %d\n", clifd);
	// read in header of GuiMsg
	curlen = 0;
	while (curlen < MTU_SIZE_UART) {
		rval = read(clifd, data, MTU_SIZE_UART);
		if (rval <= 0) {
			DEBUG_OUTPUT("Closing client socket %d.\n", clifd);
			close(clifd);
			return -1;
		} else {
			curlen += rval;
			if (rval < MTU_SIZE_UART)
				break;
		}
	}
	DEBUG_OUTPUT("reading len =============== %d.\n", curlen);
	return curlen;
}


//------------------------------------------------------------------------------
//  Command read thread
//------------------------------------------------------------------------------
void *commandReadThreadFunc(void *arg) {
	int i = 0;
	int len;
	unsigned char recvBuffer[PACKET_SIZE];
	DEBUG_OUTPUT("commandReadThread running.\n");
	
	while (1) {
		// Wait for clients
		
		// Read from clients and check for errors
		pthread_mutex_lock(&commandClientsLock);
		DEBUG_OUTPUT("commandReadThread: commandClient========== %d\n", commandClient);
		if (commandClient != -1 ) {
			DEBUG_OUTPUT("receiving.....========== %d\n", commandClient);
			len=recvCommand(commandClient, recvBuffer);
			if ( len< 0) {
				close(commandClient);
				commandClient = -1;
				DEBUG_OUTPUT("receiving.....===ERRORRRRRR===== %d\n", commandClient);
				pthread_mutex_unlock(&commandClientsLock);
			}
			pthread_mutex_unlock(&commandClientsLock);
			if(UDR[0]!=0x7e)
			{
				memcpy(UDR,recvBuffer,len);
				set_uart_int_rx(g_usart);
			}
			
		}
		else {
			
			pthread_cond_wait(&commandClientsCond, &commandClientsLock);
			pthread_mutex_unlock(&commandClientsLock);
		}
	}
	return 0;
}
